const NO_ROWS_AVAILABLE  = -1;
const NO_ROWS_SELECTED   = 0;
const SINGLE_ROW_SELECTED    = 1;
const MULTIPLE_ROWS_SELECTED = 2;

const CM_SELECT_ALL   = 0;
const CM_SELECT_NONE  = 1;

const CM_MARK_AS_READ     = 6;
const CM_MARK_ALL_AS_READ = 7;
const CM_MARK_AS_NEW  = 8;
const CM_MARK_ALL_AS_NEW  = 9;

var gFeedTree = null;
var gArticleTree = null;
var gFeedDatasource = null;
var gUpdateQueue = null;
var gPendingUpdates = null;
var gBundle;

function initFeedPanel()
{
  gFeedDatasource = mzGetNewOrCurrentDSByFilename("mzWebFeeds.rdf", true, null, false);
  gBundle = document.getElementById("multizilla-properties");

  gFeedTree = document.getElementById("FeedTree");
  gFeedTree.database.AddDataSource(gFeedDatasource);
  gFeedTree.builder.rebuild();
  gFeedTree.database.AddObserver(feedTreeRDFObserver);

  gArticleTree = document.getElementById("ArticleTree");
  gArticleTree.database.AddDataSource(gFeedDatasource);
  gArticleTree.builder.rebuild();

  restoreLastActiveChannel();
  checkRefreshInterval();

  var observerService = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
  observerService.addObserver(onLinkVisited, "link-clicked", false);
}

function checkRefreshInterval()
{
  var navigatorWindow = top.document.defaultView;
  var now = (new Date().valueOf())/1000; // seconds
  var lastRefresh = navigatorWindow.mzPreferences.readMyPref("multizilla.sidebar.feed-panel.last-refresh", "int", now);
  var pastTime = (now - lastRefresh);
  var skipRefresh = navigatorWindow.mzPreferences.readMyPref("multizilla.sidebar.feed-panel.skip-refresh", "int", 300);
  var refreshTarget = (pastTime < skipRefresh) ? ((skipRefresh - pastTime) * 1000) : 250;

  setTimeout(refreshChannels, 250); // refreshTarget);
}

function initContextMenu(aEvent)
{
  var tree = aEvent.originalTarget.parentNode;
  var treeState = getCurrentTreeState(tree);
  var menuitems = aEvent.originalTarget.childNodes;

  for (var i = 0; i < menuitems.length; i++) {
    menuitems[i].removeAttribute("disabled");
  }
  document.getElementById(tree.id + "-selectAll").setAttribute("disabled", (treeState == NO_ROWS_AVAILABLE || treeState == MULTIPLE_ROWS_SELECTED));
  document.getElementById(tree.id + "-undoSelection").setAttribute("disabled", (treeState != MULTIPLE_ROWS_SELECTED));

  if (tree.id == "FeedTree") {
    document.getElementById("show").setAttribute("disabled", (treeState == NO_ROWS_AVAILABLE || treeState == NO_ROWS_SELECTED));
    document.getElementById("add").setAttribute("disabled", (treeState == MULTIPLE_ROWS_SELECTED));
    document.getElementById("remove").setAttribute("disabled", (treeState == NO_ROWS_AVAILABLE || treeState == NO_ROWS_SELECTED));
  }
  else if (tree.id == "ArticleTree") {
    document.getElementById("openInTab").setAttribute("disabled", (treeState == NO_ROWS_AVAILABLE || treeState == NO_ROWS_SELECTED));
    document.getElementById("openInWindow").setAttribute("disabled", (treeState == NO_ROWS_AVAILABLE || treeState == NO_ROWS_SELECTED));

    var row = getRowFromEvent(tree, aEvent);
    var resource = (tree.currentIndex >= 0) ? tree.builderView.getResourceAtIndex(tree.currentIndex) : "";

    if (resource && getVisitedStatus(resource)) {
      menuitems[CM_MARK_AS_READ].setAttribute("hidden", "true");
      menuitems[CM_MARK_ALL_AS_READ].setAttribute("hidden", "true");

      menuitems[CM_MARK_AS_NEW].removeAttribute("hidden");
      menuitems[CM_MARK_ALL_AS_NEW].removeAttribute("hidden");

      menuitems[CM_MARK_AS_NEW].setAttribute("disabled", (treeState != SINGLE_ROW_SELECTED));
      menuitems[CM_MARK_ALL_AS_NEW].setAttribute("disabled", (treeState == NO_ROWS_AVAILABLE));
    }
    else {
      menuitems[CM_MARK_AS_NEW].setAttribute("hidden", "true");
      menuitems[CM_MARK_ALL_AS_NEW].setAttribute("hidden", "true");

      menuitems[CM_MARK_AS_READ].removeAttribute("hidden");
      menuitems[CM_MARK_ALL_AS_READ].removeAttribute("hidden");

      menuitems[CM_MARK_AS_READ].setAttribute("disabled", (treeState != SINGLE_ROW_SELECTED));
      menuitems[CM_MARK_ALL_AS_READ].setAttribute("disabled", (treeState == NO_ROWS_AVAILABLE));
    }
  }
  document.getElementById(tree.id + "-properties").setAttribute("disabled", (treeState != SINGLE_ROW_SELECTED));
}

function onTreeSelect(aEvent)
{
  var tree = aEvent.target;

  if (tree.id == "FeedTree") {
    if (gFeedTree.currentIndex >= 0) {
      gArticleTree.ref = gFeedTree.builderView.getResourceAtIndex(gFeedTree.currentIndex).Value;

      if (gArticleTree.ref)
        refreshVisitedStateFor(gArticleTree.ref);
    }
  }
}

function onTreeClick(aEvent)
{
  if (aEvent.button == 2 || aEvent.ctrlKey || aEvent.originalTarget.localName != "treechildren")
    return;

  var tree = aEvent.target.parentNode;
  var row = getRowFromEvent(tree, aEvent);

  if (row >= 0) {
    if (tree.id == "ArticleTree" ) {
      var target = (aEvent.ctrlKey || aEvent.button == 1) ? "newtab" : "current";
      loadArticle(row, target);
    }
    else if (aEvent.button == 1) {
      executeCommand("show", aEvent);
    }
  }
}

function onTreeCommand(aEvent)
{
  var target = (aEvent.ctrlKey) ? "newtab" : "current";
  loadArticle(gArticleTree.currentIndex, target);
}

function loadArticle(aRow, aTarget)
{
  var url = _getCellText(gArticleTree, "link", aRow);
  var navigatorWindow = top.document.defaultView;
  var useAboutBlankFirst = navigatorWindow.mzPreferences.readMyPref("multizilla.tabs.blank-tab-first", "bool", true);
  var tabbrowser = navigatorWindow.getBrowser();

  if (aTarget == "newtab") {
    if (useAboutBlankFirst)
      var browser = navigatorWindow.checkForBlankTab(tabbrowser);

    if (useAboutBlankFirst && browser)
      browser.loadURI(url);
    else
       tabbrowser.addTab(url);
  }
  else if (aTarget == "window")
     window.open(url);
  else
    tabbrowser.loadURI(url);

  var resource = gArticleTree.builderView.getResourceAtIndex(aRow);
  setStateTo(resource, "visited");
}

function doMenuCommand(aEvent)
{
  var menuID = aEvent.originalTarget.id;

  executeCommand(menuID, aEvent);
}

function executeCommand(aMenuID, aEvent)
{
  var tree = aEvent.target;
  var resourceList, resource, tabbrowser, browser, url;
  var navigatorWindow = top.document.defaultView;
  var useAboutBlankFirst = navigatorWindow.mzPreferences.readMyPref("multizilla.tabs.blank-tab-first", "bool", true);

  while (tree.localName != "tree") {
    tree = tree.parentNode;
  }

  switch(aMenuID.replace(/^\w*-/,''))
  {
     case "select-all":
     case "selectAll":
          selectRows(tree, "select-all");
          break;
     case "select-none":
     case "undoSelection":
          selectRows(tree, "select-none");
          break;
     case "show":
          tabbrowser = navigatorWindow.getBrowser();
          resourceList = getSelectedResources(tree);

          for (i in resourceList) {
            url = _getCellText(gFeedTree, "url", resourceList[i].row);

            if (useAboutBlankFirst)
              browser = navigatorWindow.checkForBlankTab(tabbrowser);

            if (useAboutBlankFirst && browser)
              browser.loadURI(url);
            else
              tabbrowser.addTab(url);
          }
          break;
     case "add":
          url = navigatorWindow.readFromClipboard();

          if (!url || !url.match(/^https?:\/\/\w.\w/))
            url = "";

          openAddFeedWindow(url);
          break;
     case "remove":
          resourceList = getSelectedResources(tree);

          if (resourceList.length) {
            var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
            var dataElement = document.getElementById("confirmationText");
            var confirmationTitle = dataElement.getAttribute("title");
            var confirmationText = dataElement.getAttribute("text");

            if (promptService.confirm(window, confirmationTitle, confirmationText)) {
              var feedRootContainer = mzConstructRDFContainer(gFeedDatasource, "seq", "urn:feeds:root", false);

              for (var i = (resourceList.length-1); i >= 0; i--) {
                resource = gFeedTree.builderView.getResourceAtIndex(resourceList[i].row);
                removeResources(gFeedDatasource, resource.Value);
                removeDataResources(gFeedDatasource, resource.Value);
                feedRootContainer.RemoveElement(resource, true);
                mzFlushDataSource(gFeedDatasource);
              }
            }
          }
          break;
     case "openInTab":
          resourceList = getSelectedResources(tree);

          for (i in resourceList) {
            loadArticle(resourceList[i].row, "newtab");
          }
          break;
     case "openInWindow":
          resourceList = getSelectedResources(tree);

          if (resourceList.length == 1) {
            loadArticle(resourceList[i].row, "window");
          }
          else {
            tabbrowser = openWindowAndReturnBrowser();

            for (i in resourceList) {
              url = _getCellText(gArticleTree, "link", resourceList[i].row);

              if (useAboutBlankFirst)
                browser = navigatorWindow.checkForBlankTab(tabbrowser);

              if (useAboutBlankFirst && browser)
                browser.loadURI(url);
              else
                tabbrowser.addTab(url);
              setStateTo(RDF.GetResource(resourceList[i].resource), "visited");
            }
          }
          break;
     case "markAllAsRead":
     case "markAllAsNew":
          if (getCurrentTreeState(tree) == SINGLE_ROW_SELECTED)
            selectRows(tree, "select-all");
     case "markAsRead":
     case "markAsNew":
          var newState = (aMenuID.search(/AsNew/) > 0) ? "unread" : "visited";
          resourceList = getSelectedResources(tree);

          for (i in resourceList) {
            setStateTo(RDF.GetResource(resourceList[i].resource), newState);
          }
          selectRows(tree, "select-none");
          break;
     case "properties":
          // resource = gFeedTree.builderView.getResourceAtIndex(tree.currentIndex);
          resource = gFeedTree.builderView.getResourceAtIndex(gFeedTree.currentIndex);
          openPropertyWindow(resource.Value);
          break;
  }
}

function openAddFeedWindow(aURL)
{
  var navigatorWindow = top.document.defaultView;

  if (aURL && navigatorWindow.mzFeedHandler.isNewFeedURL(aURL) == false) {
    var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
    var title = gBundle.getString("duplicatedFeedAlertTitle");
    var text = gBundle.getString("duplicatedFeedAlertText");
    promptService.alert(window, title, text);
  }
  else {
    var windowStyle = navigatorWindow.mzPreferences.readMyPref("multizilla.windows.classic-style", "bool", false);
    window.openDialog("chrome://multiviews/content/feedpanel/miscellaneous/addFeedWindow.xul", 
        "", "chrome, dependent, modal", window, navigatorWindow, aURL, windowStyle);
  }
}

function selectRows(aTree, aCommand)
{
  switch (aCommand) {
    case "select-all": // Called from the context menu, or Ctrl+A ?
     aTree.view.selection.selectAll();
     break;
    case "select-none": // Called from the context menu, or Ctrl+Z ?
     aTree.view.selection.clearSelection();
     aTree.view.selection.select(aTree.currentIndex);
     aTree.treeBoxObject.ensureRowIsVisible(aTree.currentIndex);
     break;
  }
  aTree.focus();
}

function openPropertyWindow(aResource)
{
  var navigatorWindow = top.document.defaultView;
  var propertyWindow = navigatorWindow.getMostRecentWindow("multizilla:feedviewer-properties", false);

  if (propertyWindow) {
    propertyWindow.focus();
  }
  else {
    var windowStyling = navigatorWindow.mzPreferences.readMyPref("multizilla.windows.classic-style", "bool", false);
    window.openDialog("chrome://multiviews/content/feedpanel/properties/propertiesWindow.xul", 
                      "", "chrome, dependent, modal", navigatorWindow, gFeedDatasource, aResource, windowStyling);
  }
}

function setStateTo(aResource, aState)
{
  if (getVisitedStatus(aResource) && aState == "") {
    var property = "http://multizilla.mozdev.org/rdf#state";
    var stateResource = gFeedDatasource.GetTarget(aResource, RDF.GetResource(property), true);
    gFeedDatasource.Unassert(aResource, RDF.GetResource(property), stateResource);
    hasNewArticles(aResource.Value, true)
  }
  else {
    mzAddOrChangeProperty(gFeedDatasource, aResource.Value, "state", aState);
    hasNewArticles(aResource.Value, true);
  }
}

// This one has to go, simply because it has been copied to feedHandler.js already!
function hasNewArticles(aResource, aUpdateChannelFlag)
{
  var state = false;
  var resource = aResource.replace(/:article-\d*/, '');
  resource = RDF.GetResource(resource);
  var channels = mzGetResourcesFromContainer(gFeedDatasource, resource);

  if (channels.length) {
    var property = "http://multizilla.mozdev.org/rdf#state";

    for (i in channels) {
      if (!gFeedDatasource.GetTarget(RDF.GetResource(channels[i]), RDF.GetResource(property), true))
        state = true;
    }
  }
  if (aUpdateChannelFlag)
    mzAddOrChangeProperty(gFeedDatasource, resource.Value, "state", (state ? "new" : ""));

  return state;
}

function refreshVisitedStateFor(aResource, aURL)
{
  var resource = RDF.GetResource(aResource);
  var articles = mzGetResourcesFromContainer(gFeedDatasource, resource);

  if (articles.length) {
    var linkProperty = "http://multizilla.mozdev.org/rdf#link";
    var stateProperty = "http://multizilla.mozdev.org/rdf#state";
    var checkAgainstHistory = (aURL == undefined);

    if (checkAgainstHistory) {
      var navigatorWindow = top.document.defaultView;
      var _nsIIOService = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
      var _nsIGlobalHistory;
      // Builds 20040210 and up (see also mozilla bug 224829)
      if ("@mozilla.org/browser/global-history;2" in Components.classes) {
    _nsIGlobalHistory = Components.classes["@mozilla.org/browser/global-history;2"].getService(Components.interfaces.nsIGlobalHistory2);
      }
      else {
    _nsIGlobalHistory = Components.classes["@mozilla.org/browser/global-history;1"].getService(Components.interfaces.nsIGlobalHistory);
      }
    }
    for (i in articles) {
      if (checkAgainstHistory) {

        var stateResource = gFeedDatasource.GetTarget(RDF.GetResource(articles[i]), RDF.GetResource(stateProperty), true);

        if (stateResource instanceof Components.interfaces.nsIRDFNode) {
          if (stateResource.QueryInterface(Components.interfaces.nsIRDFLiteral).Value == "unread")
            continue;
        }
      }
      var linkResource = gFeedDatasource.GetTarget(RDF.GetResource(articles[i]), RDF.GetResource(linkProperty), true);
      
      if (linkResource instanceof Components.interfaces.nsIRDFNode) {
        var url = linkResource.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;

        if (url) {
          if (checkAgainstHistory) {
            var uri = _nsIIOService.newURI(url, null, null);
            aURL = _nsIGlobalHistory.isVisited(uri) ? url : "";
          }
          if (url == aURL) {
            mzAddOrChangeProperty(gFeedDatasource, articles[i], "state", "visited");

            if (!checkAgainstHistory)
              break;
          }
        }
      }
    }
    hasNewArticles(aResource, true);
  }
}

function refreshChannels()
{
  var resources = mzGetResourcesFromContainer(gFeedDatasource, "urn:feeds:root");
  gPendingUpdates = new Array();

  if (resources.length) {
    var mzRDF = "http://multizilla.mozdev.org/rdf#";

    for (rIndex in resources) {
      var dataObj = new Object();
      dataObj.urn = resources[rIndex];

      var property = mzRDF + "state";
      var resource = gFeedDatasource.GetTarget(RDF.GetResource(dataObj.urn), RDF.GetResource(property), true);
      dataObj.state = resource.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;

      property = mzRDF + "url";
      resource = gFeedDatasource.GetTarget(RDF.GetResource(dataObj.urn), RDF.GetResource(property), true);
      dataObj.url = resource.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;

      property = mzRDF + "date";
      resource = gFeedDatasource.GetTarget(RDF.GetResource(dataObj.urn), RDF.GetResource(property), true);
      dataObj.date = resource.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;

      property = mzRDF + "lastModified";
      resource = gFeedDatasource.GetTarget(RDF.GetResource(dataObj.urn), RDF.GetResource(property), true);

      if (resource instanceof Components.interfaces.nsIRDFNode) {
        dataObj.lastModified = resource.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;
      }
      else {
        dataObj.lastModified = dataObj.date;
      }
      gPendingUpdates.push(dataObj);
      var now = (new Date().valueOf())/1000; // secs
      var navigatorWindow = top.document.defaultView;
      navigatorWindow.mzPreferences.writePref("int", "multizilla.sidebar.feed-panel.last-refresh", now);
    }
  }
  if (gPendingUpdates.length) {
    gUpdateQueue = new Array();
    checkPendingUpdates();
  }
}

function checkPendingUpdates()
{
  if (gPendingUpdates && gPendingUpdates.length) {
    dump("\n\nChecking for updates...");
    var dataObj = gPendingUpdates[0];
    mzAddOrChangeProperty(gFeedDatasource, dataObj.urn, "state", "checking");
    var xmlRequest = new XMLHttpRequest();
    // xmlRequest.onprogress = function(event) { onXMLProgress(event); };
    xmlRequest.open("HEAD", dataObj.url, true);

    xmlRequest.onreadystatechange = function()
    {
      if (xmlRequest.readyState == 4) {
        var lastModifiedDate = xmlRequest.getResponseHeader("Last-Modified");
        lastModifiedDate = new Date(lastModifiedDate).getTime();
        dump("\nChannel urn: " + dataObj.urn +
             "\nLast Modified date of document : " + lastModifiedDate +
             "\ndate stored in mzWebFeeds.rdf  : " + dataObj.lastModified);
        if (lastModifiedDate > dataObj.lastModified) {
          /***
            * The 'Last-Modified' send by the server is more recent, but that  
            * doesn't necessarily means that the document itself is updated,
            * but callback function updateStageTwo will take care of that.
            */
          updateStageTwo(dataObj);
        }
        else {
          hasNewArticles(dataObj.urn, true);
          /* var newState = (dataObj.state == "checking") ? "" : dataObj.state;  
          mzAddOrChangeProperty(gFeedDatasource, dataObj.urn, "state", newState); */
          removeFromPendingQueue(dataObj.urn);
        }
      }
    }
    xmlRequest.send(null);
  }
  else if (gUpdateQueue.length) {
    dump("\n\nFinalizing update process...");
    var navigatorWindow = top.document.defaultView;
    dump("\n" + gUpdateQueue.length + " in queue to refresh...");

    for (i in gUpdateQueue) {
      dump("\nRefreshing articles for: " + gUpdateQueue[i][0][8]);
      navigatorWindow.mzFeedHandler.addFeedFromDataCollection(gUpdateQueue[i], true);
    }
    gUpdateQueue = null;
  }
}

function updateStageTwo(aDataObj)
{
  dump("\nChecking date in document...aka updateStageTwo()");
  var xmlRequest = new XMLHttpRequest();
  var navigatorWindow = top.document.defaultView;

  xmlRequest.overrideMimeType("application/xml"); // "application/rss+xml"
  xmlRequest.onprogress = function(aEvent)
  {
    var request = aEvent.target;
    var url = request.channel.originalURI.spec;
    /* dump("\nQI: " + aEvent.position.QueryInterface(Components.interfaces.nsIDOMLSProgressEvent));
       dump("\naEvent   : " + aEvent +
         "\nposition : " + aEvent.position +
         "\ntotalSize: " + aEvent.totalSize); */
  }
  xmlRequest.open("GET", aDataObj.url);
  xmlRequest.setRequestHeader("User-Agent", "The MultiZilla Feed Reader/Viewer");
  xmlRequest.setRequestHeader("Cache-Control", "no-cache");
  xmlRequest.onreadystatechange = function()
  {
    if (xmlRequest.readyState == 4) {
      var channel = xmlRequest.channel.QueryInterface(Components.interfaces.nsIHttpChannel);
      var xmldoc = (channel && channel.responseStatus == 200) ? xmlRequest.responseXML : null;

      if (xmldoc) {
        var dataCollection = navigatorWindow.mzFeedHandler.getFeedInfoFromDocument(xmldoc);
        dump("\ndate inside remote document   : " + dataCollection[7] +
             "\ndate stored in mzWebFeeds.rdf : " + aDataObj.lastModified);
        if (dataCollection[7] > aDataObj.lastModified) {
          dataCollection[1] = aDataObj.url;
          dataCollection[8] = aDataObj.urn;
          dataCollection = navigatorWindow.mzFeedHandler.getDataCollectionFromDocument(xmldoc, dataCollection);
          dump("\nUpdate available for: " + aDataObj.urn);
          /***
            * We have a new update available and callback function updateStageThree
            * will remove all current items, and add this update to the update queue
            * which will later be used by callback function finalizeUpdateProcess.
            */
          updateStageThree(dataCollection);
        }
        else {
          hasNewArticles(aDataObj.urn, true);
          /* var newState = (aDataObj.state == "checking") ? "" : aDataObj.state;
          mzAddOrChangeProperty(gFeedDatasource, aDataObj.urn, "state", newState); */
          dump("\nNo update available for: " + aDataObj.urn);
         removeFromPendingQueue(aDataObj.urn);
        }
      }
    }
  }
  xmlRequest.send(null);
}

function updateStageThree(aDataCollection)
{
  var urn = aDataCollection[0][8];
  dump("\nRemoving all articles for: " + urn);
  removeResources(gFeedDatasource, urn);
  var resource = RDF.GetResource(urn);
  var channel = gFeedDatasource.ArcLabelsOut(resource);

  while (channel.hasMoreElements()) {
    var article = channel.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
    var articleResource = gFeedDatasource.GetTarget(resource, article, true);

    if (articleResource instanceof Components.interfaces.nsIRDFResource) {
      gFeedDatasource.Unassert(resource, article, articleResource);
    }
  }
  mzFlushDataSource(gFeedDatasource);
  dump("\nAdding: " + urn + " to update queue...");
  gUpdateQueue.push(aDataCollection);
  removeFromPendingQueue(urn);
}

function onXMLProgress(aEvent)
{
  // dump("\nposition : " + aEvent.position +
     //  "\ntotalSize: " + aEvent.totalSize);
}

function removeFromPendingQueue(aTargetURN)
{
  for (i in gPendingUpdates) {
    var target = (gPendingUpdates[i].urn == undefined) ? gPendingUpdates[i][8] : gPendingUpdates[i].urn;

    if (target == aTargetURN) {
      dump("\nRemoving: " + aTargetURN + " from pending queue...");
      gPendingUpdates.splice(i, 1);
    }
  }
  checkPendingUpdates();
}

function restoreLastActiveChannel()
{
  if (getCurrentTreeState(gFeedTree) != NO_ROWS_AVAILABLE) {
    var navigatorWindow = top.document.defaultView;
    var lastChannel = navigatorWindow.mzPreferences.readMyPref("multizilla.sidebar.feed-panel.last-channel", "string", "");

    if (lastChannel) {
      gArticleTree.ref = lastChannel;

      for (var i = 0; i < gFeedTree.treeBoxObject.view.rowCount; i++) {
        var resource = getResourceForRow(gFeedTree, i);

        if (resource == lastChannel) {
          gFeedTree.view.selection.select(i);
          gFeedTree.treeBoxObject.ensureRowIsVisible(i);
        }
      }
      refreshVisitedStateFor(resource);
    }
  }
}

function getVisitedStatus(aResource)
{
  var property = "http://multizilla.mozdev.org/rdf#state";
  var stateResource = gFeedDatasource.GetTarget(aResource, RDF.GetResource(property), true);

  if (stateResource instanceof Components.interfaces.nsIRDFNode) {
    if (stateResource.QueryInterface(Components.interfaces.nsIRDFLiteral).Value == "visited")
      return true;
  }
  return false;
}

function getResourceForRow(aTree, aRowNumber)
{
  return (aRowNumber > -1) ? aTree.builderView.getResourceAtIndex(aRowNumber).Value : null;
}

function getSelectedResources(aTree)
{
  var resources = new Array();
  var rows = getSelectedRows(aTree);

  for (i in rows) {
    resources[i] = { row: rows[i], 
                     resource: getResourceForRow(aTree, rows[i])
                   };
  }
  if (resources)
    return resources;
  return null;
}

function getSelectedRows(aTree)
{
  var rows = -1;
  var selectedRows = new Array();
  // Get and return all selection ranges
  var rangeCount = aTree.builderView.selection.getRangeCount();
  // Iterate over all ranges to collect the selected rows
  for (var range = 0; range < rangeCount; range++) {
    var startSelection = {};
    var endSelection = {};
    aTree.builderView.selection.getRangeAt(range, startSelection, endSelection);
    // Collect all rows and store them in our array
    for (var row = startSelection.value; row <= endSelection.value; row++) {
      // Check for invalid row numbers
      if (startSelection.value != -1 || endSelection.value != -1)
    selectedRows[++rows] = row;
    }
  }
  if (selectedRows)
    return selectedRows;
  return null;
}

function getCurrentTreeState(aTree)
{
  if (aTree.treeBoxObject.view.rowCount < 1)
    return NO_ROWS_AVAILABLE;

  var selectedRows = getSelectedRows(aTree).length;

  if (selectedRows > 1)
    return MULTIPLE_ROWS_SELECTED;
  else if (selectedRows == 1)
    return SINGLE_ROW_SELECTED;
  return NO_ROWS_SELECTED;
}

function _getCellText(aTree, aColumnID, aRow)
{
  aColumnID = ('columns' in aTree) ? aTree.columns[aColumnID] : aColumnID;
  return aTree.view.getCellText(aRow, aColumnID);
}

function getRowFromEvent(aTree, aEvent)
{
  var row = new Object();
  var col = new Object();
  var obj = new Object();

  aTree.treeBoxObject.getCellAt(aEvent.clientX, aEvent.clientY, row, col, obj);
  return row.value;
}

function shutdownFeedPanel()
{
  var channel = gArticleTree.ref;

  if (channel) {
    var navigatorWindow = top.document.defaultView;
    navigatorWindow.mzPreferences.writePref("string", "multizilla.sidebar.feed-panel.last-channel", channel);
  }
  var observerService = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
  observerService.removeObserver(onLinkVisited, "link-clicked");
  gFeedTree.database.RemoveObserver(feedTreeRDFObserver);
}

var onLinkVisited =
{
  observe: function(subject, topic, state)
  {
    if (topic == "link-clicked") {
      refreshVisitedStateFor(gArticleTree.ref, state);
    }
  }
};

var feedPanelDNDObserver =
{
  onDrop: function(aEvent, aXferData, aDragSession)
  {
     var url, data = aXferData.data.toString();
     var separator = data.indexOf("\n");

     if (separator != -1)
       url = data.substr(0, separator);
     if (url.match(/^https?:\/\/[\w.\w]/)) {
       openAddFeedWindow(url);
     }
  },

  canDrop: function(aEvent, aDragSession)
  {
     return (aEvent.target.parentNode.id == "FeedTree")
  },

  getSupportedFlavours: function ()
  {
    var flavourSet = new FlavourSet();
    flavourSet.appendFlavour("text/unicode");
    return flavourSet;
  },

  onDragStart: function(aEvent, aXferData, aDragSession) {},
  onDragOver: function(aEvent, aXferData, aDragSession) {},
  onDragExit: function(aEvent, aXferData, aDragSession) {}
};

function selectFeedAfterAdd(aResource)
{
  var row = gFeedTree.builderView.getIndexOfResource(aResource);

  if (row >= 0) {
    gFeedTree.currentIndex = row;
    gFeedTree.view.selection.clearSelection();
    gArticleTree.view.selection.clearSelection();
    gFeedTree.view.selection.select(row);
    gFeedTree.treeBoxObject.ensureRowIsVisible(row);
  }
}

var feedTreeRDFObserver =
{
  onAssert: function(aDataSource, aSource, aProperty, aTarget)
  {
    if (aSource.Value == "urn:feeds:root") {
      if (aProperty.Value.match(/#_\d*/)) {
        setTimeout(selectFeedAfterAdd, 0, aTarget);
      }
    }
  },
  onUnassert: function(aDataSource, aSource, aProperty, aTarget) {},
  onChange: function(aDataSource, aSource, aProperty, aOldTarget, aNewTarget) {},
  onBeginUpdateBatch: function() {},
  onEndUpdateBatch: function() {}
};
