/***
  * The contents of this file are subject to the Mozilla Public
  * License Version 1.1 (the "License"); you may not use this file
  * except in compliance with the License. You may obtain a copy of
  * the License at http://www.mozilla.org/MPL/
  *
  * Software distributed under the License is distributed on an "AS
  * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
  * implied. See the License for the specific language governing
  * rights and limitations under the License.
  *
  * The Original Code is this file as it was released on
  * October 30, 2004.
  *
  * The Initial Developer of the Original Code is HJ van Rantwijk.
  * Portions created by HJ van Rantwijk are Copyright (C) 2004-2005
  * HJ van Rantwijk.  All Rights Reserved.
  *
  * Contributors: 
  *   Micheal Vincent van Rantwijk <mv.van.rantwijk@gmail.com>
  *
  */
   
function findChildShell(aDocument, aDocShell, aSoughtURI)
{
  aDocShell.QueryInterface(Components.interfaces.nsIWebNavigation);
  aDocShell.QueryInterface(Components.interfaces.nsIInterfaceRequestor);
  var doc = aDocShell.getInterface(Components.interfaces.nsIDOMDocument);

  if ((aDocument && doc == aDocument) && aDocShell.currentURI.spec == aSoughtURI.spec)
    return aDocShell;

  var node = aDocShell.QueryInterface(Components.interfaces.nsIDocShellTreeNode);

  for (var i = 0; i < node.childCount; ++i) {
    var docShell = node.getChildAt(i);
    docShell = findChildShell(aDocument, docShell, aSoughtURI);

    if (docShell)
      return docShell;
  }
  return null;
}

var targetDocument = null;

var mzFeedHandler =
{
  _feedButton       : null,
  _targetBrowser    : null,
  _XSLTProcessor    : null,
  _Stylesheet       : null,
  _StylesheetURL    : "chrome://multiviews/content/feedviewer/feedviewer.xsl",
  _ATOM_10_NS       : "http://www.w3.org/2005/Atom",
  _ATOM_03_NS       : "http://purl.org/atom/ns#",
  _RSS_10_NS        : "http://purl.org/rss/1.0/",
  _RSS_09_NS        : "http://my.netscape.com/rdf/simple/0.9/",
  _RDF_NS           : "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
  _DC_NS            : "http://purl.org/dc/elements/1.1/",
  _feedViewerPath   : "chrome://multiviews/content/feedviewer/",
  _nsIIOService     : null,
  _nsIGlobalHistory : null,

  _getContentShell: function(doc, feedURI)
  {
    var browsers = getBrowser().browsers;

    for (var i = 0; i < browsers.length; i++) {
      var shell = findChildShell(doc, browsers[i].docShell, feedURI);

      if (shell)
        return { shell: shell, browser: browsers[i] };
    }
    return null;
  },
  // I copied this from Firefox but it has to go in a next release because it sucks!
  init: function()
  {
    // dump("\nmzFeedHandler.init()");
    this._nsIIOService = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
    // Builds 20040210 and up (see also mozilla bug 224829)
    if ("@mozilla.org/browser/global-history;2" in Components.classes)
      this._nsIGlobalHistory = Components.classes["@mozilla.org/browser/global-history;2"].getService(Components.interfaces.nsIGlobalHistory2);
    else
      this._nsIGlobalHistory = Components.classes["@mozilla.org/browser/global-history;1"].getService(Components.interfaces.nsIGlobalHistory);

    this._nsIIOService = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
    var self = this;

    function onPageLoad(event) {
      dump("\nmzFeedHandler.onPageLoad() triggered");
      self.checkForFeed(event);
    }
    addEventListener("load", onPageLoad, true);
  },

  checkForFeed: function(event)
  {
    var aDocument = event.originalTarget;
    // dump("\nmzFeedHandler.checkForFeed() " + aDocument);
    if (!this.validate(aDocument))
      return;

    var feedURI = Components.classes["@mozilla.org/network/io-service;1"].
                             getService(Components.interfaces.nsIIOService).
                             newURI(aDocument.location.href, null, null);

    var shellObj = this._getContentShell(aDocument, feedURI);

    if (shellObj) {
      this._targetBrowser = shellObj.browser;
      this.preProcess(feedURI.spec, event.originalTarget, "load");
    }
  },

  loadFeed: function(aBrowser, aURL)
  {
    this._targetBrowser = aBrowser;
    this.preProcess(aURL, aBrowser.contentDocument, "load");
  },

  showFeed: function(aURL, inNewTab)
  {
    // dump("\nmzFeedHandler.showFeed()");
    var tabbrowser = getBrowser();
    var tab = (inNewTab) ? tabbrowser.addTab() : tabbrowser.selectedTab;
    var browser = tabbrowser.getBrowserForTab(tab);

    this._targetBrowser = browser;
    this.getXMLDocument(aURL, "show");
  },

  validate: function(aDocument, skipCheck)
  {
    // dump("\nmzFeedHandler.validate() " + aDocument);
    if (!(aDocument instanceof XMLDocument) && !(aDocument instanceof XULDocument))
      return false;
      
    if (!this.isDocumentFeed(aDocument, false))
      return false;

    if (skipCheck == undefined) {
      var feedURI = Components.classes["@mozilla.org/network/io-service;1"].
                               getService(Components.interfaces.nsIIOService).
                               newURI(aDocument.location.href, null, null);

      if (!this._getContentShell(aDocument, feedURI))
        return false;
    }
    return true;
  },

  updateFeeds: function()
  {
    // dump("\nmzFeedHandler.updateFeeds()");
    if (!this._feedButton)
      this._feedButton = document.getElementById("feed-button");

    var feeds = gBrowser.mCurrentBrowser.feeds;

    if (!feeds || feeds.length == 0) {
      this._feedButton.removeAttribute("feeds");
      this._feedButton.setAttribute("tooltiptext", gMultiZillaBundle.GetStringFromName("feedNoFeeds"));
    } 
    else {
      this._feedButton.setAttribute("feeds", "true");
      var tooltipText = (feeds.length == 1) ? "showFeed" : "feedHasFeeds";

      if (feeds.length > 1)
        this._feedButton.setAttribute("multiple", "true");
      else
        this._feedButton.removeAttribute("multiple");

      this._feedButton.setAttribute("tooltiptext", gMultiZillaBundle.GetStringFromName(tooltipText));
    }
  },

  onFeedLinkAdded: function(aEvent)
  {
    // dump("\nzFeedHandler.onFeedLinkAdded()");
    if (!this._feedButton)
      this._feedButton = document.getElementById("feed-button");

    var erel = aEvent.target.rel;
    var etype = aEvent.target.type;
    var etitle = aEvent.target.title;

    const alternateRelRegex = /(^|\s)(alternate|\S*feed\S*)($|\s)/i;
    const titleRegex = /(^|\s)(rss|atom)($|\s)/i;
  
    if (!alternateRelRegex.test(erel) || !etype)
      return;

    etype = etype.replace(/^\s+/, "");
    etype = etype.replace(/\s+$/, "");
    etype = etype.toLowerCase();

    if (etype == "application/rss+xml" || etype == "application/atom+xml" ||
        (etype == "text/xml" || etype == "application/xml" || etype == "application/x.atom+xml" ||
        etype == "application/rdf+xml") && titleRegex.test(etitle))
    {
      const targetDoc = new XPCNativeWrapper(aEvent.target, "ownerDocument").ownerDocument;
      var feedURI = Components.classes["@mozilla.org/network/io-service;1"].
                               getService(Components.interfaces.nsIIOService).
                               newURI(targetDoc.location.href, null, null);

      var shellInfo = this._getContentShell(targetDoc, feedURI);
      var browserForLink = shellInfo.browser;

      if (!browserForLink)
        return;

      var feeds = (browserForLink.feeds != null) ? browserForLink.feeds : new Array();
      var wrapper = new XPCNativeWrapper(aEvent.target, "href", "type", "title");
      feeds.push({ href: wrapper.href, type: wrapper.type, title: wrapper.title});
      browserForLink.feeds = feeds;

      if (browserForLink == gBrowser || browserForLink == gBrowser.mCurrentBrowser) {
        this._feedButton.setAttribute("feeds", "true");
        this.updateFeeds();
      }
    }
  },

  initFeedMenu: function(aEvent)
  {
    // dump("\nzFeedHandler.initFeedMenu()");
    var feedList = gBrowser.mCurrentBrowser.feeds;

    if (feedList) {
      var menupopup = aEvent.target;

      if (feedList.length == 1) {
        this.showFeed(feedList[0].href);
      }
      else if (menupopup.parentNode.hasAttribute("multiple")) {
        deleteHistoryItems(menupopup); // this depends on sessionHistoryUI.js

        for (i in feedList) {
          var feedInfo = feedList[i];
          var menuitem = document.createElement("menuitem");
          var title = feedInfo.title || feedInfo.href;
          var label = gMultiZillaBundle.formatStringFromName("feedShowFeed", [title], 1);

          menuitem.setAttribute("label", label);
          menuitem.setAttribute("index", index);
          menuitem.setAttribute("tooltiptext", feedInfo.href);
          menuitem.setAttribute("feed", feedInfo.href);
          menupopup.appendChild(menuitem);
        }
        return true;
      }
    }
	return false;
  },

  getXMLDocument: function(aURL, aAction)
  {
    dump("\nmzFeedHandler.getXMLDocument()");
    var xmlRequest = new XMLHttpRequest();
    var handler = this;

    xmlRequest.overrideMimeType("application/xml");
    xmlRequest.onProgress = function(event) { handler.onXMLProgress(event); };
    xmlRequest.open("GET", aURL);
    xmlRequest.setRequestHeader("User-Agent", "The MultiZilla Feed Reader/Viewer");
    xmlRequest.onreadystatechange = function()
    {
      if (xmlRequest.readyState == 4) {
        var channel = xmlRequest.channel.QueryInterface(Components.interfaces.nsIHttpChannel);
        var xmldoc = (channel && channel.responseStatus == 200) ? xmlRequest.responseXML : null;

        if (xmldoc)
          handler.preProcess(aURL, xmldoc, aAction);
      }
    }
    xmlRequest.send(null);
  },

  onXMLProgress: function(aEvent)
  {
    dump("\nposition : " + aEvent.position +
         "\ntotalSize: " + aEvent.totalSize);
  },

  preProcess: function(aURL, aXMLDocument, aAction)
  {
    // dump("\nmzFeedHandler.preProcess() " + aAction);
    switch(aAction) {
      case "load":
      case "show":
          if (this.isDocumentFeed(aXMLDocument, false)) {
            this._XSLTProcessor = new XSLTProcessor();
            this._Stylesheet = document.implementation.createDocument("", "", null);
            this._process(this.transformFeed, aXMLDocument, aAction);
            mzAddToHistory(aURL);
          }
          break;
      case "subscribe":
           if (this.isDocumentFeed(aXMLDocument, false)) {
             var dataCollection = this.getDataCollectionFromDocument(aXMLDocument);

             if (dataCollection) {
               dataCollection[0][1] = aURL;
               this.addFeedFromDataCollection(dataCollection, false);
             }
           }
           else {
             alert("Sorry, this is not a valid feed!");
           }
           break;
    }
  },

  _process: function(callback, aXMLDocument, aAction)
  {
    // dump("\nmzFeedHandler._process()");
    var handler = this;

      function onStyleSheetLoaded() {
        handler._Stylesheet.removeEventListener("load", onStyleSheetLoaded, false);
        handler._XSLTProcessor.importStylesheet(handler._Stylesheet);
        callback(aXMLDocument, aAction);
      }

    this._Stylesheet.addEventListener("load", onStyleSheetLoaded, false);
    this._Stylesheet.load(this._StylesheetURL);
  },

  doCleanup: function()
  {
    // dump("\nmzFeedHandler.doCleanup()");
    this._targetDocument = null;
    this._XSLTProcessor = null;
    this._Stylesheet = null;
  },

  isDocumentFeed: function(aDocument, aReturnObjFlag)
  {
    // dump("\nmzFeedHandler.isDocumentFeed() " + aDocument);
    aDocument = aDocument.documentElement;

    var rootName = aDocument.localName;
    // dump("\nrootName: " + rootName);
    var rootNS = aDocument.namespaceURI;
    dump("\nrootNS: " + rootNS);
    if (aReturnObjFlag)
      var feedInfo = null;

    if (rootName == "feed" && (rootNS == this._ATOM_10_NS || rootNS == this._ATOM_03_NS)) {
      if (aReturnObjFlag) {
        feedInfo = new Object();
        feedInfo.type = rootName;
        feedInfo.version = aDocument.getAttribute("version") || "undetermined";
        feedInfo.namespace = rootNS;
      }
      else {
        return true;
      }
    }
    else {
      var channel;

      if (rootName == "rss" && rootNS == null) {
        channel = aDocument.getElementsByTagName('channel')[0];
      }
      else if (rootName == "RDF" && rootNS == this._RDF_NS) {
        channel = aDocument.getElementsByTagNameNS(this._RDF_NS, 'channel')[0] ||
                  /* aDocument.getElementsByTagNameNS(this._RSS_09_NS, 'channel')[0] || */
                  aDocument.getElementsByTagNameNS(this._RSS_10_NS, 'channel')[0];
      }
      if (channel && channel.parentNode == aDocument) {
        if (aReturnObjFlag) {
          feedInfo = new Object();
          feedInfo.type = rootName;
          feedInfo.version = aDocument.getAttribute("version") || "undetermined";
          feedInfo.namespace = rootNS;
        }
        else {
          return true;
        }
      }
    }
    return (aReturnObjFlag) ? feedInfo : false;
  },

  isDocumentTransformedFeed: function(aDocument)
  {
    // dump("\nmzFeedHandler.isDocumentTransformedFeed()");
    var isTransformed = false;
    /***
      * The given document might already been 'XSL Transformed' so we need to check this
      * Note: The check in Mozilla Firefox fails to detect these documents as feed!
      */
    var linkElements = aDocument.getElementsByTagName('link');
 
    if (linkElements.length) {
      // There are at least two link elements, but there might be more!
      for (var i = 0; i < linkElements.length; i++) {
        if (linkElements[i].getAttribute('rel') == "stylesheet" &&
            linkElements[i].getAttribute('href') == this._feedViewerPath + "feedviewer.css")
        {
          isTransformed = true;
          break;
        }
      }
      /***
        * This should be an XSLTransformed document, or someone is totally nuts (because 
        * he used an invalid link to a chrome: stylsheet) but let's be absolutely sure
        * about it and also check for our <script> element.
        */
      if (isTransformed) {
        var scriptElement = aDocument.getElementsByTagName('script')[0];

        if (scriptElement && scriptElement.getAttribute('type') == "application/x-javascript" &&
            scriptElement.getAttribute('src') == this._feedViewerPath + "feedviewer.js")
        {
          return 2;
        }
      }
    }
    return this.validateFeed(aDocument.documentElement);
  },

  getFeedDescription: function(aDocument, aTitle)
  {
    // dump("\nzFeedHandler.getFeedDescription()");
    var docElement = aDocument.documentElement;
    var rootName = docElement.localName;
    var rootNS = docElement.namespaceURI;
    var description;
  
    if (rootName == "feed" && (rootNS == this._ATOM_10_NS || rootNS == this._ATOM_03_NS))
      description = docElement.getElementsByTagNameNS(rootNS, 'tagline')[0];
    else if (rootName == "rss" && rootNS == null)
      description = docElement.getElementsByTagName('description')[0];
    else if (rootName == "RDF" && rootNS == this._RDF_NS)
      description = docElement.getElementsByTagNameNS(this._RDF_NS, 'description')[0];

    if (description)
      return description.textContent.length ? description.textContent : aTitle;
    else
    {
      /***
        * XSL Transformed documents will have a <meta> element which contains the description
        * Note: this will only be checked when this.isDocumentTransformedFeed() returned 2 !!
        */
      if ((aDocument instanceof HTMLDocument || aDocument instanceof XMLDocument || 
          aDocument instanceof XULDocument) && this.isDocumentTransformedFeed(docElement) == 2)
      {
        var metaElement = docElement.getElementsByTagName('meta');

        for (var i = 0; i < metaElement.length; i++) {
          description = (metaElement[i].getAttribute('name') == 'Description' && 
                         metaElement[i].hasAttribute('content'))

          if (description) {
            description = metaElement[i].getAttribute('content');
            return description.length ? description: aTitle;
          }
        }
      }
    }
    return aTitle;
  },

  transformFeed: function(aXMLDocument, aAction)
  {
    // dump("\nmzHandler.transformFeed() " + aAction);
    var handler = mzFeedHandler;
    var xmldoc = aXMLDocument;
    var processor = handler._XSLTProcessor;
    var initAfterPageLoad = (aAction == "load");

    processor.setParameter(null, "url", xmldoc.documentURI);
    processor.setParameter(null, "title", gMultiZillaBundle.GetStringFromName("feedTitle"));
    processor.setParameter(null, "titleHeader", gMultiZillaBundle.GetStringFromName("feedTitleHeader"));

    var description = xmldoc.getElementsByTagName("description")[0] ||
                      xmldoc.getElementsByTagName("tagline")[0] ||
                      xmldoc.getElementsByTagName("title")[0];

    processor.setParameter(null, "description", description.textContent);

    var articleCount = xmldoc.getElementsByTagName("item").length ||
                       xmldoc.getElementsByTagName("entry").length;

    processor.setParameter(null, "articleCount", gMultiZillaBundle.formatStringFromName("feedDescription", [articleCount], 1));
    processor.setParameter(null, "currentBuildDate", new Date().toUTCString());
    processor.setParameter(null, "versionInfo", gMultiZillaBundle.GetStringFromName("feedSoftware"));
    // if () {
      processor.setParameter(null, "subscribeButtonText", "Subscribe to feed");
    // }
    var targetDocument;

    if (initAfterPageLoad)
      targetDocument = document.implementation.createDocument("", "", null);
    else 
      targetDocument = handler._targetBrowser.contentDocument;

    var newFragment = processor.transformToFragment(xmldoc, targetDocument);

    targetDocument = (initAfterPageLoad) ? xmldoc : targetDocument;

    _documentElement = targetDocument.documentElement;
    // Remove everything up to the document's <head> element
    while (_documentElement.hasChildNodes()) {
      _documentElement.removeChild(_documentElement.firstChild);
    }
    if (initAfterPageLoad) {
      _documentElement.appendChild(newFragment);
    }
    else {
      var newNodes = new Array();
      // Collect <head> and <body> elements from our fragment
      for (var i = 0; i < newFragment.firstChild.childNodes.length; i++) {
        newNodes.push(newFragment.firstChild.childNodes[i]);
      }
      // Append the new elements into the target document
      for (i in newNodes) {
        _documentElement.appendChild(newNodes[i]);
      }
    }
    // Get the title
    var h1s = targetDocument.getElementsByTagName("h1");
    // Do we have a title?
    if (h1s.length) 
      targetDocument.title = h1s[0].textContent;

    var subscribeButton = targetDocument.getElementById("subscribeButton");

    if (handler.isNewFeedURL(xmldoc.documentURI)) {
      subscribeButton.addEventListener("click", function(event){handler.subscribeFromButton(event.target, xmldoc.documentURI);}, true);
    }
    else {
      subscribeButton.style.display = "none";
    }
    handler.doCleanup();
  },

  isNewFeedURL: function(aURL)
  {
    dump("\nChecking " + aURL + " in isNewFeedURL()");
    var _nsIIOService = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
    aURL = aURL.replace(/www\./,'');
    var uri = _nsIIOService.newURI(aURL, null, null);
    var urn = "urn:" + uri.host + ":feed-1";

    var feedDatasource = mzGetNewOrCurrentDSByFilename("mzWebFeeds.rdf", true, null, false);

    if (checkResourceForValue(feedDatasource, "urn:feeds:root", urn)) {
      dump("\nChecking " + urn + " because of a possible match...");
      var channels = mzGetResourcesFromContainer(feedDatasource, "urn:feeds:root");

      if (channels.length) {
        var property = "http://multizilla.mozdev.org/rdf#url";

        for (i in channels) {
          var urlResource = feedDatasource.GetTarget(RDF.GetResource(channels[i]), RDF.GetResource(property), true);
          // We might want to remove the replace() after the testers fixed their RDF!
          var url = urlResource.QueryInterface(Components.interfaces.nsIRDFLiteral).Value.replace(/www\./,'');

          if (url == aURL) {
            dump("\nThis feed is already listed!");
            return false;
          }
        }
      }
    }
    dump("\nThis feed isn't filed yet!");
    return true;
  },


  showFeedInBrowser: function(aURL, aBrowser)
  {
    // dump("\nmzHandler.showFeedInBrowser");
    this._targetBrowser = aBrowser;
    this.getXMLDocument(aURL, "show");
  },

  isFeedLoad: function(aURL, aBrowser)
  {
    // dump("\nmzHandler.isLoadingAFeed: " + aURL);
    var request = new XMLHttpRequest();
    var mimeTypeRegEx = /(^|\s)(application|text)\/(xml|rss\+xml|atom\+xml|rdf\+xml)($|\s)/i

    request.open("HEAD", aURL, true);
    request.onreadystatechange = function()
    {
      if (request.readyState == 4) {
        var contentType = request.getResponseHeader("Content-Type");

        if (mimeTypeRegEx.test(contentType)) {
          dump("\nIS FEED: " + contentType);
          aBrowser.isFeed = true;
          // aBrowser.stop();
          // mzFeedHandler.showFeedInBrowser(aURL, aBrowser);
          // aBrowser.currentURI.spec = aURL;
        }
      }
    }
    request.send(null);
  },

  getFeedInfoFromDocument: function(aDocument)
  {
    // dump("\nmzHandler.getFeedInfoFromDocument()" + aDocument);
    var rootNode = aDocument.documentElement;
    var feedInfo = new Array();
    var version = rootNode.getAttribute("version");
    feedInfo[0] = rootNode.localName;
    feedInfo[0] += (version) ? " " + rootNode.getAttribute("version") : "";
    feedInfo[1] = ""; // url
    feedInfo[2] = ""; // favicon
    var element = aDocument.getElementsByTagName("title")[0];
    feedInfo[3] = element.textContent;
    // dump("\nfeedInfo[3]: " + feedInfo[3]);
    element = aDocument.getElementsByTagName("link")[0];
    feedInfo[4] = (rootNode.localName == "feed") ? element.getAttribute("href") : element.textContent;
    // dump("\nfeedInfo[4]: " + feedInfo[4]);
    element = aDocument.getElementsByTagName("description")[0] ||
              aDocument.getElementsByTagName("tagline")[0];
    feedInfo[5] = element.textContent;
    // dump("\nfeedInfo[5]: " + feedInfo[5]);
    element = aDocument.getElementsByTagName("image")[0];
    var imageAtributes = [ "", "", "", "" ];

    if (element) {
      var nodes = aDocument.createTreeWalker(element, 1, null, true);
      var node = nodes.currentNode;

      while (node != null) {
        var _nodeName = node.nodeName;

        if (_nodeName == "url")
          imageAtributes[0] = node.textContent;
        else if (_nodeName == "title")
          imageAtributes[1] = node.textContent;
        else if (_nodeName == "width")
          imageAtributes[2] = node.textContent;
        else if (_nodeName == "height")
          imageAtributes[3] = node.textContent;

        node = nodes.nextNode();
      }
    }
    feedInfo[6] = imageAtributes;
    // dump("\nfeedInfo[6]: " + feedInfo[6]);
    element = aDocument.getElementsByTagName("pubDate")[0] ||
              aDocument.getElementsByTagName("lastBuildDate")[0] ||
              aDocument.getElementsByTagName("modified")[0] ||
              aDocument.getElementsByTagName("issued")[0] ||
              aDocument.getElementsByTagName("updated")[0] ||
              aDocument.getElementsByTagNameNS(this._DC_NS, "date")[0];
    if (element) {
      var date = element.textContent;

      if (date.search(/^\d\d\d\d/) != -1)
        date = this._W3CToIETFDate(date);
      feedInfo[7] = new Date(date.replace(/BST|IST$/, "GMT +0100")).getTime();
    }
    else {
      feedInfo[7] = "";
    }
    dump("\nfeedInfo[7]: " + feedInfo[7]);
    /***
      * Please don't break the callers by adding new items without taking care of them!
      */
    return feedInfo;
  },

  getDataCollectionFromDocument: function(aDocument, aDataCollection)
  {
    // dump("\nmzHandler.getDataCollectionFromDocument() " + aDocument);
    var rootNode = aDocument.documentElement;
    var rootNS = rootNode.namespaceURI;
    var articleTag = (rootNode.localName == "feed") ? "entry" : "item"
    var articles = aDocument.getElementsByTagName(articleTag);
    var dataCollection = new Array();
    dataCollection[0] = (aDataCollection == undefined) ? this.getFeedInfoFromDocument(aDocument) : aDataCollection;

    for (var i = 0; i < articles.length; i++) {
      var nodes = aDocument.createTreeWalker(articles[i], 1, null, true);
      var node = nodes.currentNode;
      var articleInfo = new Object();
      var collectedNodes = 0;

      while (node != null) {
        var titleFromLink = "";
        var articleLink = "";
        var _nodeName = node.nodeName;

        if (_nodeName == "title") {
          articleInfo.title = node.textContent;
          collectedNodes++;
        }
        else if (_nodeName == "link") {
          var rel = node.getAttribute("rel");

          if (rel && rel.match(/^service./)) {
            // skip service.edit
          }
          else {
            articleLink = (articleTag == "entry") ? node.getAttribute("href") : node.textContent;
            articleInfo.url = (articleLink) ? articleLink : "";
            titleFromLink = node.getAttribute("title");
            collectedNodes++;
          }
        }
        else if (_nodeName == "updated" || _nodeName == "modified" || _nodeName == "published" || 
                 _nodeName == "pubDate" || _nodeName == "dc:date" || _nodeName == "issued")
        {
          var date = node.textContent;
          // Do we have 4 digits at the start i.e. "2004-09-13T09:54:04-08:00"
          if (date.search(/^\d\d\d\d/) != -1)
            date = this._W3CToIETFDate(date);
          articleInfo.published = new Date(date).getTime();
          collectedNodes++;
        }
        if (articleInfo.title == undefined && titleFromLink)
          articleInfo.title = titleFromLink;
        if (collectedNodes == 3) // Bail out as early as we can!
          node = null;
        else
          node = nodes.nextNode();
      }
      dataCollection.push(articleInfo);
    }
    return (dataCollection.length >= 1) ? dataCollection : null;
  },

  addFeedFromDataCollection: function(aDataCollection, aRefreshFlag)
  {
    // dump("\nmzHandler.addFeedFromDataCollection() " + aDataCollection);
    var feedDatasource = mzGetNewOrCurrentDSByFilename("mzWebFeeds.rdf", true, null, false);
    var feedRootContainer = mzConstructRDFContainer(feedDatasource, "seq", "urn:feeds:root", false);

    if (!feedRootContainer)
      return;

    var i = 1;
    var attribute = 0;
    var url = aDataCollection[0][1].replace(/www\./,'');
    var mzRDF = "http://multizilla.mozdev.org/rdf#";
    var feedContainer, property, uri, urn, orgURN;

    if (aRefreshFlag == undefined || aRefreshFlag == false) {
      uri = this._nsIIOService.newURI(url, null, null);
      urn = "urn:" + uri.host + ":feed-1";
      orgURN = "urn:" + uri.host;

      var value;
      var tabbrowser = getBrowser();
      aDataCollection[0][2] = tabbrowser.buildFavIconString(uri);

      while(checkResourceForValue(feedDatasource, "urn:feeds:root", urn)) {
        urn = orgURN + ":feed-" + String(i++);
      }
      feedContainer = mzConstructRDFContainer(feedDatasource, "seq", urn, false);

      var propertyNames = [ "feed", "url", "image", "title", "link", "description", "imageUrl", 
                            "imageTitle", "imageWidth", "imageHeight", "date" ];

      for (i = 0; i < 6; i++) {
        property = mzRDF + propertyNames[i];
        value = aDataCollection[0][i];

        if (value)
          feedDatasource.Assert(RDF.GetResource(urn), RDF.GetResource(property), RDF.GetLiteral(value), true);
      }
      for (i = 6; i < 10; i++) {
        property = mzRDF + propertyNames[i];
        value = aDataCollection[0][6][attribute++];

        if (value)
          feedDatasource.Assert(RDF.GetResource(urn), RDF.GetResource(property), RDF.GetLiteral(value), true);
      }
      property = mzRDF + propertyNames[10];
      feedDatasource.Assert(RDF.GetResource(urn), RDF.GetResource(property), RDF.GetLiteral(aDataCollection[0][7]), true);
    }
    else {
      urn = aDataCollection[0][8];
      feedContainer = mzConstructRDFContainer(feedDatasource, "seq", urn, false);
    }
    // Add article entries
    for (i = 1; i < aDataCollection.length; i++) {
      var article = RDF.GetResource(urn + ":article-" + String(i));
      property = mzRDF + "title";
      var title = (aDataCollection[i].title == undefined) ? "No Title Provided" : aDataCollection[i].title;
      feedDatasource.Assert(article, RDF.GetResource(property), RDF.GetLiteral(title), true);
      property = mzRDF + "link";
      feedDatasource.Assert(article, RDF.GetResource(property), RDF.GetLiteral(aDataCollection[i].url), true);
      var date;

      if (aDataCollection[i].published) {
        date = aDataCollection[i].published;
        property = mzRDF + "published";
        feedDatasource.Assert(article, RDF.GetResource(property), RDF.GetLiteral(date), true);
      }
      else {
        date = aDataCollection[0][7];
      }
      if (date) {
        property = mzRDF + "prettydate";
        date = new Date(date).toUTCString();
        var prettyDate = date.substr(5, 11);
        feedDatasource.Assert(article, RDF.GetResource(property), RDF.GetLiteral(prettyDate), true);
      }
      uri = this._nsIIOService.newURI(aDataCollection[i].url, null, null);

      if (this._nsIGlobalHistory.isVisited(uri)) {
        property = mzRDF + "state";
        feedDatasource.Assert(article, RDF.GetResource(property), RDF.GetLiteral("visited"), true);        
      } 
      feedContainer.AppendElement(article);
    }
    if (!checkResourceForValue(feedDatasource, "urn:feeds:root", urn)) {
      feedRootContainer.AppendElement(RDF.GetResource(urn));
    }
    mzAddOrChangeProperty(feedDatasource, urn, "lastModified", String(aDataCollection[0][7]));
    this.hasNewArticles(feedDatasource, urn, true);
    // mzAddOrChangeProperty(feedDatasource, urn, "state", "new");
    mzFlushDataSource(feedDatasource);
  },

  hasNewArticles: function(aDataSource, aResource, aUpdateChannelFlag)
  {
    var state = false;
    var resource = aResource.replace(/:article-\d*/, '');
    resource = RDF.GetResource(resource);
    var channels = mzGetResourcesFromContainer(aDataSource, resource);

    if (channels.length) {
      var property = "http://multizilla.mozdev.org/rdf#state";

      for (i in channels) {
        if (!aDataSource.GetTarget(RDF.GetResource(channels[i]), RDF.GetResource(property), true))
          state = true;
      }
    }
    if (aUpdateChannelFlag) {
      mzAddOrChangeProperty(aDataSource, resource.Value, "state", (state ? "new" : ""));
    }
    return state;
  },

  subscribeToFeed: function(aURL)
  {
    this.getXMLDocument(aURL, "subscribe");
  },

  subscribeFromButton: function(aTarget, aURL)
  {
    while (aTarget.localName != "div") {
      aTarget = aTarget.parentNode;
    }
    aTarget.style.display = "none";
    this.openAddFeedWindow(aURL);
  },

  openAddFeedWindow: function(aURL)
  {
    if (aURL && this.isNewFeedURL(aURL) == false) {
      var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
      var title = gBundle.getString("duplicatedFeedAlertTitle");
      var text = gBundle.getString("duplicatedFeedAlertText");
      promptService.alert(window, title, text);
    }
    else {
      var windowStyle = mzPreferences.readMyPref("multizilla.windows.classic-style", "bool", false);
      window.openDialog("chrome://multiviews/content/feedpanel/miscellaneous/addFeedWindow.xul", 
                        "", "chrome, dependent, modal", window, top.document.defaultView, aURL, windowStyle);
    }
  },

  getFeedLastModifiedData: function(aURL)
  {
    var request = new XMLHttpRequest();
    request.open("HEAD", aURL, true);
    request.onreadystatechange = function()
    {
      if (request.readyState == 4) {
        return request.getResponseHeader("Last-Modified");
      }
    }
    request.send(null);
  },

  _W3CToIETFDate: function(dateString)
  {  
   /* const HOURS_TO_MINUTES = 60;
    const MINUTES_TO_SECONDS = 60;
    const SECONDS_TO_MILLISECONDS = 1000;
    const MINUTES_TO_MILLISECONDS = MINUTES_TO_SECONDS * SECONDS_TO_MILLISECONDS;
    const HOURS_TO_MILLISECONDS = HOURS_TO_MINUTES * MINUTES_TO_MILLISECONDS; */
    var parts = dateString.match(/(\d\d\d\d)(-(\d\d))?(-(\d\d))?(T(\d\d):(\d\d)(:(\d\d)(\.(\d+))?)?(Z|([+-])(\d\d):(\d\d))?)?/);
    var date = new Date(parts[1], parts[3]-1, parts[5], parts[7] || 0, parts[8] || 0, parts[10] || 0, parts[12] || 0);
    var remoteToUTCOffset = 0;

    if (parts[13] && parts[13] != "Z") {
      var direction = (parts[14] == "+" ? 1 : -1);
      if (parts[15])
        remoteToUTCOffset += direction * parts[15] * 3600000; // HOURS_TO_MILLISECONDS
      if (parts[16])
        remoteToUTCOffset += direction * parts[16] * 60000; // MINUTES_TO_MILLISECONDS
    }
    remoteToUTCOffset = remoteToUTCOffset * -1;
    var UTCToLocalOffset = date.getTimezoneOffset() * 60000; // MINUTES_TO_MILLISECONDS
    UTCToLocalOffset = UTCToLocalOffset * -1;
    date.setTime(date.getTime() + remoteToUTCOffset + UTCToLocalOffset);
    return date.toUTCString();
    /* scriptable date formater, for pretty printing dates
    var dateService = window.Components.classes["@mozilla.org/intl/scriptabledateformat;1"].getService(Components.interfaces.nsIScriptableDateFormat);
    return (date.valueOf()) ? dateService.FormatDateTime("", dateService.dateFormatLong, dateService.timeFormatSeconds, date.getFullYear(), date.getMonth()+1, date.getDate(), date.getHours(), date.getMinutes(), date.getSeconds()) : unknown; */
  }
};
