/* 
   - The contents of this file are subject to the Mozilla Public
   - License Version 1.1 (the "License"); you may not use this file
   - except in compliance with the License. You may obtain a copy of
   - the License at http://www.mozilla.org/MPL/
   -
   - Software distributed under the License is distributed on an "AS
   - IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
   - implied. See the License for the specific language governing
   - rights and limitations under the License.
   -
   - The Original Code is this file as it was released on
   - October 4, 2003.
   -
   - The Initial Developer of the Original Code is HJ van Rantwijk.
   - Portions created by HJ van Rantwijk are Copyright (C) 2003-2005
   - HJ van Rantwijk.  All Rights Reserved.
   -
*/

const ADDBUTTON    = 0;
const MERGEBUTTON  = 1;
const RELOADBUTTON = 3;

const NO_ROWS_AVAILABLE      = -1;
const NO_ROWS_SELECTED       = 0;
const MIXED_ROWS_SELECTED    = 1;
const MULTIPLE_ROWS_SELECTED = 2;
const TAB_ROW_SELECTED       = 3;
const WINDOW_ROW_SELECTED    = 4;

const TITLE_PROPERTY    = 0;
const LOCATION_PROPERTY = 4;
const SELECTED_PROPERTY = 6;
const ORDINAL_PROPERTY  = 7;
const TABSTYLE_PROPERTY = 13;

const nsIWindowDataSource = Components.interfaces.nsIWindowDataSource; // We need this for multiviewsRDF.js!
const nsIRDFResource = Components.interfaces.nsIRDFResource;
const nsIRDFLiteral = Components.interfaces.nsIRDFLiteral;

var mzPreferences;

var gBrowser = null; // declared in navigator.js but out of reach for us!

var gResource = "";
var bundle;
var gSelectedWindowID = null;
var contextMenuitems = null;

var prefWindow = null;
var prefPanelNames;
var managerTypes = [ "sessionManager" ];
var dialogType = "sessionManagerPrefWindow";
var panelPath = "chrome://multiviews/content/session-manager/preferences/";
var gUseStackedWindows = false;
// Global vars used by confirmation dialog
var gConfirmCheckbox = false;
var gEscapeKeyPressed = false;
var gConfirmForID = 0;

/***
  * Toolbar button ID arrays that will be replaced with elements in updateToolbarButtons()
  * note: keep the order in-line with the corresponding XUL code!
  */
var toolbarButtons = [ "addSessionButton", "mergeTabsButton", "openSessionButton", "reloadButton", "propertiesButton", "renameButton", "deleteButton", "prefsButton" ];
/***
  * Toolbar button arrays, to ease the pain of enabling/disabling the buttons
  */
var defaultEnabledButtons      = [  true, false, false, false, false, false, false, true ];
var tabEnabledButtons          = [  true, false,  true,  true,  true, false,  true, true ];
var windowEnabledButtons       = [  true, false,  true, false, false,  true,  true, true ];
var selectionEnabledButtons    = [  true, false,  true, false, false, false,  true, true ];
var multipleRowsEnabledButtons = [  true, false,  true, false, false, false,  true, true ];

////////////////////////////////////////////////////////////////////////////////
function selectWindowStyling(aWindow)
{
  var documentElement = aWindow.document.documentElement;

  if (readMyPref("multizilla.windows.classic-style", "bool", false))
    documentElement.removeAttribute("extraStyling");
  else
    documentElement.setAttribute("extraStyling", "true");
}
////////////////////////////////////////////////////////////////////////////////
function readMyPref(aPrefString, aPrefType, aDefaultValue)
{
  var navigatorWindow = getNavigatorWindow(true);
  mzPreferences = navigatorWindow.mzPreferences;
  return navigatorWindow.mzPreferences.readMyPref(aPrefString, aPrefType, aDefaultValue);
}
////////////////////////////////////////////////////////////////////////////////
function writePref(aPrefType, aPrefString, aValue)
{
  var navigatorWindow = getNavigatorWindow(true);
  mzPreferences = navigatorWindow.mzPreferences;
  return navigatorWindow.mzPreferences.writePref(aPrefType, aPrefString, aValue);
}
////////////////////////////////////////////////////////////////////////////////
function getTabbrowser(aOpenFlag)
{
  var navigatorWindow = getNavigatorWindow(aOpenFlag);

  if (navigatorWindow)
    return navigatorWindow.getBrowser();
  return null;
}
////////////////////////////////////////////////////////////////////////////////
function getBrowser()
{
  return getTabbrowser(true);
}
////////////////////////////////////////////////////////////////////////////////
// This function is copied over from multiviewsInit.js because it is not 
// available from multiviewsRDF.js (read 'it is not included').
////////////////////////////////////////////////////////////////////////////////
function getNumberOfWindowsByType(aType)
{
  var defaultType = "navigator:browser";

  if (!aType) // default check for browser window
    aType = defaultType;

  var windowCounter = 0;
  var windowManager = Components.classes['@mozilla.org/appshell/window-mediator;1'].getService(Components.interfaces.nsIWindowMediator);
  var enumerator = windowManager.getEnumerator(aType);

  while (enumerator.hasMoreElements()) {
    var aWindow = enumerator.getNext();
    windowCounter++;
  }
  return windowCounter;
}
////////////////////////////////////////////////////////////////////////////////
function updateToolbarButtons()
{
  // Check if the toolbar buttons are initialized
  if (typeof toolbarButtons[0] == 'string') {
    // Not initialized yet, initialize button array
    for (var i = 0; i < toolbarButtons.length; i++)
      toolbarButtons[i] = document.getElementById(toolbarButtons[i]);
  }
  var buttons = defaultEnabledButtons;
  var treeState = sessionManager.getCurrentTreeState();

  if (treeState > NO_ROWS_SELECTED) {

    // No, check tree selection state and choose the appropriate toolbar button array
    if (treeState == TAB_ROW_SELECTED)
      buttons = tabEnabledButtons;
    else if (treeState == WINDOW_ROW_SELECTED)
      buttons = windowEnabledButtons;
    else if (treeState == MIXED_ROWS_SELECTED)
      buttons = selectionEnabledButtons;
    else if (treeState == MULTIPLE_ROWS_SELECTED)
      buttons = multipleRowsEnabledButtons;
  }
  // Enable/disable the toolbar buttons based on the toolbar array
  for (var buttonIndex = 0; buttonIndex < toolbarButtons.length; buttonIndex++)
    toolbarButtons[buttonIndex].disabled = !buttons[buttonIndex];

  toolbarButtons[ADDBUTTON].disabled = areAllTabsBlank(tabbrowser);
  toolbarButtons[MERGEBUTTON].disabled = areAllTabsBlank(tabbrowser) || treeState == NO_ROWS_SELECTED || treeState != WINDOW_ROW_SELECTED;
}
////////////////////////////////////////////////////////////////////////////////
function initContextMenu(aEvent)
{
  tabbrowser = getTabbrowser(true);

  var i = 0;
  // Disable all menu items, but the first one, for clicking on an empty spot in the tree
  if (sessionManager.getRowNumber(aEvent) == -1) {
    var menuItems = aEvent.originalTarget.childNodes;
    // Iterate over all menu items and disable them
    for (i = 1; i < menuItems.length; i++)
      menuItems[i].setAttribute("disabled", true);
  }
  var rows = tree.treeBoxObject.view.rowCount;
  var selectedResources = sessionManager.getSelectedResources();
  // Disable "Select All" when there's a selection
  var selectAll = document.getElementById("SelectAll");
  // Disable "Deselect All" when there's no selection
  var deselectAll = document.getElementById("UndoSelection");

  if (selectedResources.length > 1) {
    selectAll.setAttribute("disabled", "true");
    deselectAll.removeAttribute("disabled");
  }
  else { // No selection, enable "Select All" and disable "Deselect All"
    if (rows > 1)
      selectAll.removeAttribute("disabled");
    deselectAll.setAttribute("disabled", "true");
  }
  var tabElements = document.getElementsByAttribute("display_for", "tabs");
  var windowElements = document.getElementsByAttribute("display_for", "windows");
  var tabAndWindowElements = document.getElementsByAttribute("display_for", "tabs-and-windows");
  var isMixed = sessionManager.checkForMixedSelection();
  var isContainer = isMixed ? false : tree.treeBoxObject.view.isContainer(sessionManager.getRowNumber(aEvent));
  var isWindow = isMixed ? false : isContainer;
  var isTab = isMixed ? false : !isWindow;

  for (i = 0; i < tabElements.length; i++) {
    tabElements[i].setAttribute("disabled", !isTab);
    tabElements[i].hidden = !isTab;
  }
  for (i = 0; i < windowElements.length; i++) {
    windowElements[i].setAttribute("disabled", !isWindow);
    windowElements[i].hidden = !isWindow;
  }
  for (i = 0; i < tabAndWindowElements.length; i++) {
    tabAndWindowElements[i].setAttribute("disabled", !isMixed);
    tabAndWindowElements[i].hidden = !isMixed;
  }
  document.getElementById('FileAsBookmark').setAttribute("disabled", (isContainer || selectedResources.length > 1));
  document.getElementById('FileAsGroupmark').setAttribute("disabled", (!isContainer && selectedResources.length == 1));
  document.getElementById('SetInterval').setAttribute("disabled", (!isTab || selectedResources.length > 1));
  document.getElementById('AddNewSession').setAttribute("disabled", areAllTabsBlank(tabbrowser));
  document.getElementById('MergeTabs').setAttribute("disabled", areAllTabsBlank(tabbrowser) || (selectedResources.length == 0) || (!isContainer));
  document.getElementById('OpenSMPropertyWindow').setAttribute("disabled", (isContainer || selectedResources.length > 1));
  document.getElementById('RenameSession').setAttribute("disabled", (!isContainer || selectedResources.length > 1));
}
//////////////////////////////////////////////////////////////////////////////
function initReloadPopup(aPopup, aEvent)
{
  if (toolbarButtons[RELOADBUTTON].disabled) {
    aEvent.preventDefault();
    return;
  }
  tabbrowser.initReloadPopup(aPopup, null, sessionManager.getReloadInterval());
}
//////////////////////////////////////////////////////////////////////////////
function openSMPropertyWindow()
{
  var treeState = sessionManager.getCurrentTreeState();

  if (treeState == NO_ROWS_SELECTED || treeState == MULTIPLE_ROWS_SELECTED || currentRow == -1)
    return;
  gResource = sessionManager.getResourceForRow(tree, currentRow); // tab    -> urn:session-x:window-x:tabs:tab-x
                                                            // window -> urn:session-x:window-x:tabs:root
  var winFeatures = "chrome, dependent, modal, dialog" + (gUseStackedWindows ? "" : ", resizable=yes, centerscreen");
  window.openDialog("chrome://multiviews/content/session-manager/properties/sessionManagerProperties.xul", 
                    "", winFeatures, window);
}
//////////////////////////////////////////////////////////////////////////////
function openPrefOverlay()
{
  var winFeatures = "chrome, dependent, modal, dialog" + (gUseStackedWindows ? "" : ", resizable=yes, centerscreen");
  window.openDialog("chrome://multiviews/content/session-manager/preferences/sessionManagerPrefWindow.xul", 
                    "", winFeatures, window, "");
}
//////////////////////////////////////////////////////////////////////////////
function getTargetWindow(aTitle, aButtonText)
{
  gSelectedWindowID = null;

  window.openDialog("chrome://multiviews/content/session-manager/miscellaneous/sessionManagerSelectWindow.xul", "",
                    "chrome, dependent, modal, dialog, centerscreen", aTitle, aButtonText);

  if (gSelectedWindowID) {
    var windowManagerDS = Components.classes['@mozilla.org/rdf/datasource;1?name=window-mediator'].getService(nsIWindowDataSource);
    return windowManagerDS.getWindowForResource(gSelectedWindowID);
  }
  return null;
}
//////////////////////////////////////////////////////////////////////////////
function areAllTabsBlank(aTabbrowser)
{
  if (!aTabbrowser)
    aTabbrowser = getTabbrowser(true);

  var numberOfTabs = aTabbrowser.mTabContainer.childNodes.length;
  var numberOfBlankTabs = getNumberOfBlankTabs(aTabbrowser);

  return (numberOfTabs == numberOfBlankTabs);
}
//////////////////////////////////////////////////////////////////////////////
function getNumberOfBlankTabs(aTabBrowser)
{
  var blankTabs = 0;
  var browsers = aTabBrowser.mPanelContainer.childNodes.length;

  for (var i = 0; i < browsers; i++) {
    var browser = aTabBrowser.getBrowserAtIndex(i);
    var tab = aTabBrowser.getTabForBrowser(browser);

    if (tab.hasAttribute("busy") || 'currentURI' in browser == false) {
      if (tab.getAttribute("firstPage") == "about:blank")
        blankTabs++;
    }
    else if (browser.currentURI.spec == "about:blank")
      blankTabs++;
  }
  return blankTabs;
}
//////////////////////////////////////////////////////////////////////////////
// This is a dummy function, called from the pref manager (for compatibility only)
function switchTree(gPanel)
{
  return;
}
//////////////////////////////////////////////////////////////////////////////
function switchPanel(aPanel)
{
  var categoryButton = null;
  var panelFile = null;

  prefWindow.gSecondLevelPrefs = false;

  if (aPanel == "sessionManagerPrefWindow")
    panelFile = "sessionManagerMainPanel.xul";
  prefWindow.refreshPanel(panelFile, categoryButton);
}
//////////////////////////////////////////////////////////////////////////////
function initWindowOpenType()
{
  gUseStackedWindows = readMyPref("multizilla.managers.use-stacked-windows", "bool", /Win/.test(navigator.platform));
}
////////////////////////////////////////////////////////////////////////////////
function getConfirmation(aWindow, aTitle, aText, aCheckboxText)
{
  gConfirmCheckbox = false;
  var checkbox = {};
  var checkboxText = aCheckboxText ? stringBundle.getString(aCheckboxText) : null;
  var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
  var result = promptService.confirmEx(aWindow, stringBundle.getString(aTitle), stringBundle.getString(aText),
                                       (promptService.BUTTON_POS_0 * promptService.BUTTON_TITLE_YES) + 
                                       (promptService.BUTTON_POS_1 * promptService.BUTTON_TITLE_NO),
                                       null, null, null, checkboxText, checkbox);

  gConfirmCheckbox = checkbox.value;
  return (result == 0); // Yes returns 0 / No, Esc and X returns 1
}
//////////////////////////////////////////////////////////////////////////////
function bookmarkSelection()
{
  var selectedResources = sessionManager.getSelectedResources();
  // Do we have any resources?
  if (selectedResources) {
    var groupmarkInfo = new Array();
    var mainResources = new Array();
    var selectedResourceInfo = { name: "", url: "", charset: null };
    var propertyValues, i = -1;
    // Iterate over all resources
    for (rIndex in selectedResources) {
      var resource = selectedResources[rIndex];       
      // Is this a session container?
      if (resource.match(/root/) == "root") {
        // Yes, get all resources
        var resourceList = mzGetResourcesFromContainer(mzRDFStore, resource);
        // Iterate over all container items
        for (cIndex in resourceList) {
          resource = resourceList[cIndex];
          i++;
          propertyValues = getTabDataFromResource(mzRDFStore, RDF.GetResource(resource));
          groupmarkInfo[i] = new Object();
          groupmarkInfo[i].name = propertyValues[TITLE_PROPERTY];
          groupmarkInfo[i].url = propertyValues[LOCATION_PROPERTY];
          /* We need to store the character set when we save sessions!
             groupmarkInfo[i].charset = propertyValues[CHARSET_PROPERTY]; */
          groupmarkInfo[i].charset = window.opener.document.characterSet;

          if (propertyValues[SELECTED_PROPERTY] == "true")
            selectedResourceInfo = groupmarkInfo[i];
        }
      }
      else { // Not a session container
        i++;
        propertyValues = getTabDataFromResource(mzRDFStore, RDF.GetResource(resource));
        groupmarkInfo[i] = new Object();
        groupmarkInfo[i].name = propertyValues[TITLE_PROPERTY];
        groupmarkInfo[i].url = propertyValues[LOCATION_PROPERTY];
        /* We need to store the character set when we save sessions!
           groupmarkInfo[i].charset = propertyValues[CHARSET_PROPERTY]; */
        groupmarkInfo[i].charset = window.opener.document.characterSet;

        if (propertyValues[SELECTED_PROPERTY] == "true")
          selectedResourceInfo = groupmarkInfo[i];
      }
    }
  }
  var type = (selectedResources.length > 1) ? "addGroup,group" : "";
  openDialog("chrome://communicator/content/bookmarks/addBookmark.xul", "",
             "chrome, dependent, modal, dialog, resizable=yes, centerscreen",
             selectedResourceInfo.name, selectedResourceInfo.url, null, 
             selectedResourceInfo.charset, type, groupmarkInfo);
}
//////////////////////////////////////////////////////////////////////////////
var sessionManager = {

  tree : null,
  tabbrowser : null,
  lastResource : null,
  reloadInterval : 0,
  currentRow : null,
  // tooltipBox : null,
  // tabNumberTooltip : null,
  // sessionNumberTooltip : null,
  // windowNumberTooltip : null,
  stringBundle : null,
  dragResourceList : null,

  initSessionManager: function()
  {
    tree = document.getElementById("sessionManagerTree");
    initializeRDFStorageFile();
    tree.database.AddDataSource(mzRDFStore);
    tree.builder.rebuild();
    tree.builder.addObserver(sessionTreeBuilderObserver);
    sessionTreeBuilderObserver._init();
    tabbrowser = getTabbrowser(true);
    gBrowser = getBrowser();
    stringBundle = document.getElementById("properties");
    // tooltipBox = document.getElementById("tooltipBox");
    // tabNumberTooltip = tooltipBox.childNodes[0];
    // sessionNumberTooltip = tooltipBox.childNodes[1];
    // windowNumberTooltip = tooltipBox.childNodes[2];
    lastResource = null;
    reloadInterval = 0;
    currentRow = -1;

    if (!window.arguments || (window.arguments && window.arguments[1] == 'undefined' || window.arguments[1] == false)) {
      // No need to open a useless Tab Session Manager!
      if (tree.treeBoxObject.view.rowCount == 0) {
        var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
        promptService.alert(window, stringBundle.getString("mzWizard"), stringBundle.getString("tsmNoSessionsAvailable"));
        window.close();
      }
    }
    prefPanelNames = stringBundle.getString("prefPanelNames").split(',');
    sessionManager.setDefaultButton();
    updateToolbarButtons();
    sizeToContent();
    initWindowOpenType();

    if (window.arguments && window.arguments[0] == "preferences")
      setTimeout(openPrefOverlay, 0);
  },

  reInitSessionManager: function()
  {
    tree = document.getElementById("sessionManagerTree");
    tree.builder.rebuild();
    lastResource = null;
  },

  getRowNumber: function(aEvent, aTree)
  {
    if (aTree == undefined)
      aTree = tree;

    var X = aEvent.pageX;
    var Y = aEvent.pageY;
    return aTree.treeBoxObject.getRowAt(X, Y);
  },

  getResourceForRow: function(aTree, aRowNumber)
  {
    if (aTree == undefined)
      aTree = tree;

    return (aRowNumber > -1) ? aTree.builderView.getResourceAtIndex(aRowNumber).Value : null;
  },

  getResourceSpecs: function(aResource)
  {
    var resourceValues = aResource.replace(/^urn:session-/, '').replace(/[a-z:.#\/]/g, '').split('-');
    return resourceValues;
  },

  getSelectedResources: function(aTree)
  {
    if (aTree == undefined)
      aTree = tree;

    var resources = new Array();
    var rows = this.getSelectedRows(aTree);

    for (i in rows) {
      var resource = this.getResourceForRow(aTree, rows[i]);

      if (resource)
        resources.push(resource);
    }
    if (resources.length)
      return resources;
    return null;
  },

  getSelectedRows: function(aTree)
  {
    if (aTree == undefined)
      aTree = tree;

    var rows = -1;
    var selectedRows = new Array();
    // Get and return all selection ranges
    var rangeCount = aTree.view.selection.getRangeCount();
    // Iterate over all ranges to collect the selected rows
    for (var range = 0; range < rangeCount; range++) {
      var startSelection = new Object();
      var endSelection = new Object();
      aTree.view.selection.getRangeAt(range, startSelection, endSelection);
      // Collect all rows and store them in our array
      for (var row = startSelection.value; row <= endSelection.value; row++) {
        // Check for invalid row numbers
        if (startSelection.value != -1 || endSelection.value != -1)
          selectedRows[++rows] = row;
      }
    }
    if (selectedRows)
      return selectedRows;
    return null;
  },

  getCurrentTreeState: function()
  {
    if (tree.treeBoxObject.view.rowCount < 1)
      return NO_ROWS_AVAILABLE;

    var selectedRows = sessionManager.getSelectedRows();

    if (selectedRows.length == 0)
      return NO_ROWS_SELECTED;

    var isMixed = sessionManager.checkForMixedSelection();

    if (isMixed)
      return MIXED_ROWS_SELECTED;

    var isMultiple = (selectedRows.length > 1);

    if (isMultiple)
      return MULTIPLE_ROWS_SELECTED;

    var isContainer = (currentRow > -1) ? tree.treeBoxObject.view.isContainer(tree.currentIndex) : false;
    var isWindow = isMultiple ? false : isContainer;
    var isTab = isMultiple ? false : !isWindow;

    if (isTab)
      return TAB_ROW_SELECTED;
    else if (isWindow)
      return WINDOW_ROW_SELECTED;
    return NO_ROWS_SELECTED;
  },

  getReloadInterval: function(aEvent)
  {
    var interval = 0;
    var columnID = "sm-interval";
    columnID = ('columns' in tree) ? tree.columns[columnID] : columnID;

    if (aEvent) {
      var row = this.getRowNumber(aEvent);
      interval = (row > -1) ? tree.view.getCellText(row, columnID) : 0;
    }
    else if (!aEvent && tree.currentIndex != -1)
      interval = tree.view.getCellText(tree.currentIndex, columnID);

    if (interval)
      return interval;
    return 1;
  },

  setReloadInterval: function(aEvent, aValue)
  {
    var resource = this.getResourceForRow(tree, currentRow);
    var intervalProp = RDF.GetResource("http://multizilla.mozdev.org/rdf#interval");
    var currentValue = mzRDFStore.GetTarget(RDF.GetResource(resource), intervalProp, true);

    if (aValue == 0) // Remove interval property from tree
      mzRDFStore.Unassert(RDF.GetResource(resource), intervalProp, currentValue);
    else
      addOrChangeValue(mzRDFStore, RDF.GetResource(resource), intervalProp, currentValue, RDF.GetLiteral(aValue));
    mzSetModifiedDate(mzRDFStore, resource);
  },

  /* disableIntervals: function(aTabResource)
  {
    var containerResource = aTabResource.replace(/tab-\w* /, 'root'); !!!note the extra space!!!
    var mzRDF = "http://multizilla.mozdev.org/rdf#";
    var property = mzRDF + "interval";

    mzRemoveProperties(containerResource, [property]);
  }, */

  setRowTooltip: function(aEvent)
  {
    var row = this.getRowNumber(aEvent);
    if (row == -1)
      return;

    var resource = this.getResourceForRow(tree, row);
    var resourceValues = this.getResourceSpecs(resource);

    if (resourceValues.length == 3) {
      tabNumberTooltip.setAttribute("value", "Tab: " + Number(resourceValues[2]));
      tabNumberTooltip.removeAttribute("hidden");
    }
    else
      tabNumberTooltip.setAttribute("hidden", "true");

    windowNumberTooltip.setAttribute("value", "Window: " + Number(resourceValues[1]));
    sessionNumberTooltip.setAttribute("value", "Session: " + Number(resourceValues[0]));
  },

  setDefaultButton: function()
  {
    var dialogButton = document.documentElement.getButton("accept");
    dialogButton.removeAttribute("default");
    dialogButton = document.documentElement.getButton("cancel");
    dialogButton.setAttribute("default", "true");
    dialogButton.focus();
  },

  convertResourceDataToNumbers: function(aDataArray)
  {
     for (i in aDataArray)
       aDataArray[i] = Number(aDataArray[i])+1;
     return aDataArray;
  },

  checkForMixedSelection: function()
  {
    var tabs = 0;
    var windows = 0;
    var mixed = false;

    // Collect selected tree rows
    var selectedItems = this.getSelectedRows();
    // Do we have any selected items?
    if (selectedItems && selectedItems.length > 0) {
      // Iterate over the seleted items
      for (var row = 0; row < selectedItems.length; row++) {
        if (tree.treeBoxObject.view.isContainer(selectedItems[row]))
          windows++;
        else
          tabs++;
        // Do we have a mixed selection?
        if (tabs > 0 && windows > 0) {
          mixed = true;
          break;
        }
      }
    }
    return mixed;
  },

  removeWindowItemsFromSelection: function()
  {
    // Collect selected tree rows
    var selectedItems = this.getSelectedRows();
    // Do we have any selected items?
    if (selectedItems) {
      // Iterate over the seleted items to clear the window items
      for (row in selectedItems) {
        if (tree.treeBoxObject.view.isContainer(row))
          tree.view.selection.clearRange(row, row);
      }
    }
  },

  selectRows: function(aTree, aCommand)
  {
    if (aTree == undefined)
      aTree = tree;

    switch (aCommand) {
      case "all": // Called from the context menu, or Ctrl+A ?
           aTree.view.selection.selectAll();
           // Only remove window items when we're calle for the main tree
           if (aTree.id == "sessionManagerTree")
             this.removeWindowItemsFromSelection();
           break;
      case "none": // Called from the context menu, or Ctrl+Z ?
           aTree.view.selection.clearSelection();

           if (aTree.currentIndex >= 0)
             aTree.view.selection.select(aTree.currentIndex);
           break;
    }
    aTree.focus();
  },

  renameTabSession: function()
  {
    if (tree.currentIndex != -1) {
      var resource = sessionManager.getResourceForRow(tree, tree.currentIndex);

      if (resource) {
        var aWindow = getNavigatorWindow();
        if (!aWindow) {
          aWindow = getNavigatorWindow(true);
          setTimeout(sessionManager.renameTabSession, 0);
          return;
        }
        else 
          aWindow = window;
        renameTabSession(mzRDFStore, resource, aWindow);
      }
    }
  },

  addNewSession: function()
  {
    var browserWindow, windowID, sessionID = 0;

    if (getNumberOfWindowsByType() > 1) {
      var title = stringBundle.getString("addSessionTitle");
      var buttonText = stringBundle.getString("addSessionButton");
      browserWindow = getTargetWindow(title, buttonText);
    }
    else
      browserWindow = getNavigatorWindow(true);
    if (!browserWindow)
      return;
    windowID = getRelativeWindowResourceID(browserWindow);
    sessionID = getNextSessionID(sessionID, windowID);

    var selectedTabStyle = readMyPref("multizilla.session-manager.selectedtab-styling", "string", "bold");
    var resource = "urn:" + sessionID + ":" + windowID + ":tabs:root";
    storeGroupData(mzRDFStore, resource, browserWindow.gBrowser, false, false, selectedTabStyle);
    tree.builder.rebuild();
  },

  mergeTabs: function()
  {
    var browserWindow, resource = this.getResourceForRow(tree, currentRow);
    resource = mzGetLastResourceFromContainer(mzRDFStore, resource);
    var resourceValues = this.getResourceSpecs(resource);
    var tabIndex = Number(resourceValues[2]) + 1;

    if (getNumberOfWindowsByType() > 1) {
      var title = stringBundle.getString("mergeTabsTitle");
      var buttonText = stringBundle.getString("mergeTabsButton");
      browserWindow = getTargetWindow(title, buttonText);
    }
    else
      browserWindow = getNavigatorWindow(true);

    if (!browserWindow)
      return;

    var historyItemsToKeep = readMyPref("multizilla.tabs.save-session.history-items", "int", 0);
    addTabDataSeqToDS(mzRDFStore, resource, tabIndex, browserWindow.gBrowser, true, "", (historyItemsToKeep > 0), true);
    mzSetModifiedDate(mzRDFStore, resource);
  },

  doConfirm: function(additionalText)
  {
    var dialogTitle = stringBundle.getString("tsmConfirmActionTitle");
    var dialogText = stringBundle.getString("tsmConfirmActionText");
    additionalText = stringBundle.getString(additionalText);
    dialogText = dialogText.replace(/%S/, additionalText);
    var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
                                  .getService(Components.interfaces.nsIPromptService);
    var result = promptService.confirmEx(window, dialogTitle, dialogText,
                                         (promptService.BUTTON_POS_0 * promptService.BUTTON_TITLE_YES) + 
                                         (promptService.BUTTON_POS_1 * promptService.BUTTON_TITLE_NO),
                                         null, null, null, null, {value:0});
    tree.focus();
    return (result == 0); // Yes returns 0 / No, Esc and X returns 1
  },

  onClick: function(aEvent)
  {
    if (aEvent.detail == 2)
      return;

    this.doClick(aEvent);
  },

  onDblClick: function(aEvent)
  {
    if (aEvent.button == 1)
      return;

    this.doClick(aEvent);
  },

  doClick: function (aEvent)
  {
    currentRow = this.getRowNumber(aEvent);

    if (currentRow == -1 || !tree.view.selection.isSelected(currentRow))
      return;

    var isContainer = tree.treeBoxObject.view.isContainer(currentRow);
    reloadInterval = sessionManager.getReloadInterval(aEvent);

    if (aEvent.button == 0 && aEvent.detail == 2 && !isContainer || aEvent.button == 1) {
      if (isContainer && aEvent.ctrlKey) {
        if (this.getSelectedResources() == null) {
          tree.currentIndex = currentRow;
          tree.view.selection.clearSelection();
          tree.view.selection.select(currentRow);
          tree.treeBoxObject.ensureRowIsVisible(currentRow);
        }
        sessionManager.onAccept(true, false, -1);
        return;
      }
      sessionManager.onAccept(true, false, false);
    }
  },

  onSelect: function()
  {
    currentRow = tree.currentIndex;
    document.documentElement.getButton("accept").removeAttribute("disabled");
    updateToolbarButtons();
  },

  onRemove: function(skipLastResourceCheck)
  {
    var rows = this.getSelectedRows();
    var confirmActionText = (rows.length > 1) ? "tsmRemoveMultipleItems" : "tsmRemoveSingleItem";

    if (!this.doConfirm(confirmActionText))
      return;

    if (rows && rows.length) {
      var containersToBeValidated = new Array();
      // Delete the selected row(s) from the bottom to the top
      for (var i = rows.length-1; i > -1; i--) {
        /***
          * 'urn:session-x:window-x:tabs:root' for windows (containers)
          * 'urn:session-x:window-x:tabs:tab-x' for tabs (normal rows)
          */
        var resource = this.getResourceForRow(tree, rows[i]);
        var containerResource = resource.replace(/tab-\w*/, 'root');

        if (containersToBeValidated.length) {
          for (cIndex in containersToBeValidated) {
            if (containerResource != containersToBeValidated[cIndex])
              containersToBeValidated.push(containerResource);
          }
        }
        else
          containersToBeValidated.push(containerResource);
        // Prevent duplicated actions!
        if (!skipLastResourceCheck && (lastResource == null || lastResource == resource))
          break;
        if (tree.treeBoxObject.view.isContainer(rows[i])) // Window
          removeAllResourcesFor(mzRDFStore, resource);
        else { // Single Tab
          removeTabResourcesFor(mzRDFStore, resource);
        }
      }
      for (cIndex in containersToBeValidated) {
        containerResource = containersToBeValidated[cIndex];
        sessionTreeBuilderObserver._checkContainerIndex(containerResource);
        sessionTreeBuilderObserver._validateContainer(containerResource);
        mzSetModifiedDate(mzRDFStore, containerResource);
      }
    }
    this.reInitSessionManager();
  },

  onHelp: function(aPageAnchor)
  {
    var pageAnchor = "";
    var helpPages = stringBundle.getString("helpPages-SessionManager");

    if (aPageAnchor) {
      if (typeof aPageAnchor == 'string')
        pageAnchor = "#" + aPageAnchor;
      else // event
        pageAnchor = "#" + aPageAnchor.target.id.replace(/^sessionManager-/, '');
    }
    // We could eventually check the on-line state!
    tabbrowser = getTabbrowser(true);
    tabbrowser.selectedTab = tabbrowser.addOrGetTab(helpPages + pageAnchor);
    window.close();
  },

  shutdown: function()
  {
    tree.builder.removeObserver(sessionTreeBuilderObserver);
  },

  onCancel: function()
  {
    var retValue = this.confirmClose(gConfirmForID);

    if (retValue)
      this.shutdown();

    return retValue;
  },

  onAccept: function(skipLastResourceCheck, closeSessionManager, openInNewWindow)
  {
    var rows = this.getSelectedRows();
    var result;

    if (rows && rows.length) {
      mzTabIndex = 0;
      var startupFlag = false;
      // [Open] openInNewWindow = true [click] openInNewWindow = false!
      closeSessionManager = (openInNewWindow == true); // 'openInNewWindow' can be 0, 1 or -1
      // check for open browser window
      if (getNavigatorWindow(false) == null) { // no open browser window
        tabbrowser = getTabbrowser(true); // open a new browser window and re-initialize 'tabbrowser'
        openInNewWindow = false;
        startupFlag = true;
      }
      for (i in rows) {
        var resource = this.getResourceForRow(tree, rows[i]);
        // Prevent duplicated actions!
        if (!skipLastResourceCheck && (/* lastResource == null || */ lastResource == resource)) {
          continue;
        }
        if (tree.treeBoxObject.view.isContainer(rows[i])) { // Restore Window(s)
          // Do we really need to open a new window?
          openInNewWindow = (openInNewWindow >= 0) ? !areAllTabsBlank(tabbrowser) : false;
          // Open new browser window if the user clicked on the Ok button, or 
          // when he tries to fool us, by closing all available browser windows!
          var browser = openInNewWindow ? openWindowAndReturnBrowser() : tabbrowser;
          result = restoreTabsFromResource(mzRDFStore, browser, resource, openInNewWindow);
        }
        else { // Restore Tab(s)
          var newTab = (mzTabIndex > 0 || rows.length == 1);
          startupFlag = newTab;
          mzTabIndex++;
          result = restoreTabFromResource(mzRDFStore, tabbrowser, RDF.GetResource(resource), newTab, startupFlag);
        }
        if (result == -1)
          break;
      }
    }
    if (closeSessionManager) {
      this.shutdown();
      window.close();
    }
  },

  confirmClose: function(aConfirm)
  {
    /***
      * 1  = Close Button
      * 2  = ESC key pressed
      * 4  = Ctrl-W
      * 8  = Click on [X] on Window
      */
    var shouldClose = true;
    const confirmPrefID = "multizilla.session-manager.confirm.quit";
    var confirmPrefValue = readMyPref(confirmPrefID, "int", 15);
    // Is the pref set for confirmation?
    if (confirmPrefValue && (confirmPrefValue & aConfirm) == aConfirm) {
      const checkboxText = 'managerSkipConfirmFor';
      var dialogCheckboxText;
      // Yes, ask the users confirmation
      switch(aConfirm) {
        case (aConfirm & 1): dialogCheckboxText = checkboxText + 'CloseButton';
          break;
        case (aConfirm & 2): dialogCheckboxText = checkboxText + 'EscapeKey';
          break;
        case (aConfirm & 4): dialogCheckboxText = checkboxText + 'CtrlW';
          break;
        case (aConfirm & 8): dialogCheckboxText = checkboxText + 'WindowX';
          break;
      }
      shouldClose = getConfirmation(window, 'tmConfirmCloseTitle', 'tsmConfirmCloseText', dialogCheckboxText);
      // Is the checkbox checked and do we have a pref value to work with?
      if (gConfirmCheckbox && (confirmPrefValue >= aConfirm)) {
        // Yes, update pref
        writePref("int", confirmPrefID, (confirmPrefValue -= aConfirm));
      }
    }
    return shouldClose;
  },

  refreshTabStyling: function(aNewStyle)
  {
    var rows = tree.treeBoxObject.view.rowCount;

    for (var row = 0; row < rows; row++)
    {
      if (tree.treeBoxObject.view.isContainer(row)) {
        var windowResource = sessionManager.getResourceForRow(tree, row);
        var windowContainer = mzRDFCUtils.MakeSeq(mzRDFStore, RDF.GetResource(windowResource)).GetElements();

        while (windowContainer.hasMoreElements())
        {
          var tabResource = windowContainer.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
          var tabContainer = mzRDFCUtils.MakeSeq(mzRDFStore, RDF.GetResource(tabResource)).GetElements();
          var labels = mzRDFStore.ArcLabelsOut(tabResource);

          while (labels.hasMoreElements())
          {
            try {
              var label = labels.getNext().QueryInterface(Components.interfaces.nsIRDFResource);

              if (label.Value == "http://multizilla.mozdev.org/rdf#selected") {
                var data = mzRDFStore.GetTarget(tabResource, label, true);
                var value = data.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;

                if (value == "true")
                  addOrChangeStyleProperty(mzRDFStore, tabResource.Value, aNewStyle);
                else
                  mzRDFStore.Unassert(tabResource, label, data);
              }
            } 
            catch(ex) { 
              // die silently
            }
          }
	    }
      }
    }
  },

  sortSessionContainers: function(aDataSource)
  {
    var sessionContainer = mzRDFCUtils.MakeSeq(aDataSource, RDF.GetResource("urn:sessions:root")).GetElements();
    var resourceList = new Array();

    while (sessionContainer.hasMoreElements()) {
      var containerResource = sessionContainer.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
      resourceList.push(containerResource.Value);
    }
    if (resourceList.length) {
      function compare(a, b)
      {
        if (a < b) return -1;
        if (a > b) return 1;
        return 0;
      }
      resourceList = resourceList.sort(compare);

      for (rIndex in resourceList) {
        var resource = resourceList[rIndex];
        dump("\nvalidating: " + resource);
        // sessionTreeBuilderObserver._validateContainer(RDF.GetResource(resource));
      }
    }
  },

  validateSessionContainers: function(aDataSource)
  {
    var sessionContainer = mzRDFCUtils.MakeSeq(aDataSource, RDF.GetResource("urn:sessions:root")).GetElements();
    var resourceList = new Array();

    while (sessionContainer.hasMoreElements()) {
      var containerResource = sessionContainer.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
      resourceList.push(containerResource);
    }
    if (resourceList.length) {
      for (rIndex in resourceList) {
        var resource = resourceList[rIndex];
        dump("\nvalidating: " + resource.Value);
        sessionTreeBuilderObserver._validateContainer(resource);
      }
    }
  },

  onDragStart: function(aEvent, aXferData, aDragSession)
  {
    if (tree.treeBoxObject.view.rowCount > 0) {
      if (aEvent.ctrlKey)
        aDragSession.action = Components.interfaces.nsIDragService.DRAGDROP_ACTION_COPY;

      sessionTreeBuilderObserver.dragSession = aDragSession;
      var X = aEvent.pageX;
      var Y = aEvent.pageY;
      var dragStartRow = tree.treeBoxObject.getRowAt(X, Y);

      if (dragStartRow == -1 || tree.treeBoxObject.view.isContainer(dragStartRow))
        return;

      var startResource = this.getResourceForRow(tree, dragStartRow);
      var resource, resources = this.getSelectedResources();
      // Do we have any selected rows?
      if (resources == null) {
        // No, and this happens when you start a d&d copy action without first selecting (a) row(s)
        resources = [startResource];
        this.dragResourceList = [startResource];
      }
      else {
        var alreadyAdded = false;
        this.dragResourceList = resources;

        for (rIndex in resources) {
          resource = resources[rIndex];
          if (resource.match(/tabs:root$/))
            return;
          if (startResource == resource) {
            alreadyAdded = true;
            break;
          }
        }
        if (!alreadyAdded)
          return;
          // this.dragResourceList.push(resources[i]);
      }
      var resourceList = "";

      for (i in resources) {
        resource = resources[i];
        resourceList = resourceList.concat((i == 0 || i == resources.length) ? resource : "," + resource);
      }
      aXferData.data = new TransferData();
      aXferData.data.addDataForFlavour("text/x-moz-url", "sm:resources:" + resourceList);
    }
  },

  onDragExit: function(aEvent, aXferData, aDragSession)
  {
    this.dragResourceList = null;
  },

  canDrop: function (aEvent, aDragSession)              {},
  onDragOver: function(aEvent, aXferData, aDragSession) {},
  onDrop: function(aEvent, aXferData, aDragSession)     {},
  getSupportedFlavours: function ()                     {}

};

/***
  * Drag & Drop support for the Session Manager (v1.6.5.0)
  * written by HJ on a rainy, but hot, Saterday afternoon.
  *
  * Notes: I've only tested it with my Mozilla 1.7 build.
  *        Should now also work with Mozilla 1.8 builds!
  */
var sessionTreeBuilderObserver =
{
  DROP_BEFORE : -1, // Components.interfaces.nsITreeView.DROP_BEFORE,
  DROP_ON     : 0,  // Components.interfaces.nsITreeView.DROP_ON,
  DROP_AFTER  : 1,  // Components.interfaces.nsITreeView.DROP_AFTER,

  dragSession     : null,
  dragTarget      : "",
  dragAction      : null,
  dragTargetIndex : -1,

  _init: function()
  {
     /***
       * Preserve backward compatibility with older Mozilla versions/builds
       * FYI: the values in older builds are: 1, 2, and 3
       */
     if ('inDropBefore' in Components.interfaces.nsITreeView) { // Old Mozilla version?
       this.DROP_BEFORE = 0;
       this.DROP_ON     = 1;
       this.DROP_AFTER  = 2;
     }
  },

  _prepareDragAction: function(aResource, aDragAction, aContainerFlag)
  {
     if (aDragAction == "append" && aContainerFlag)
       aResource = aResource.replace(/tab-\w*/, 'root');

     this.dragTarget = aResource;
     this.dragAction = aDragAction;
     return true;
  },

  _isDropOnParentContainer: function(aTargetResource, aOrientation)
  {
    var targetSessionResource = (aTargetResource.match(/root/) == "root") ? aTargetResource : aTargetResource.replace(/tab-\w*/, 'root');
    var sessionContainer = mzRDFCUtils.MakeSeq(mzRDFStore, RDF.GetResource(targetSessionResource));
    var containerItems = sessionContainer.GetCount();

    for (i in sessionManager.dragResourceList) {
      // convert 'urn:session-N:window-N:tabs:tab-N' to 'urn:session-N:window-N:tabs:root'
      var resource = sessionManager.dragResourceList[i].replace(/tab-\w*/, 'root');
      // prevent invalid/unsupported drops
      if (aTargetResource == resource || 
          containerItems == 1 && targetSessionResource == resource && aOrientation != this.DROP_ON)
        return this._prepareDragAction("", null, -1);
    }
    return false;
  },

  _checkContainerIndex: function(aTargetContainerResource)
  {
    var resourceList = mzGetResourcesFromContainer(mzRDFStore, RDF.GetResource(aTargetContainerResource));
    var newTargetResource = aTargetContainerResource.replace(/tabs:root/, 'tabs:tab-');
    // Iterate over the container items
    for (i in resourceList) {
      var resource = resourceList[i];
      var targetResource = newTargetResource + Number(i);
      /***
        * The 'resource' and 'targetResource' should normally be equal, but they will 
        * be different in older versions of MultiZilla when you delete one or more tabs 
        * from session containers, so that is the reason why we need this extra check!
        */
      if (resource != targetResource) {
        sessionManager.dragResourceList = [resource];
        /***
          * We might be called after a tab has been removed!
          * Note that we have to simulate a Drag and Drop (move) action 
          * or onDrop(insert) will fail with a JavaScript error!
          */
        if (this.dragSession == null)
          this.dragSession = {action: Components.interfaces.nsIDragService.DRAGDROP_ACTION_MOVE};

        this.dragAction = "insert";
        this.dragTarget = resource;
        this.onDrop(-1, null); // Emulate drag and drop action
        return -1;
      }
    }
    return true;
  },

  _validateContainer: function (aTargetContainerResource)
  {
    var resourceList = mzGetResourcesFromContainer(mzRDFStore, RDF.GetResource(aTargetContainerResource));
    var resource, currentValue, selectedTab = -1;
    var selectedTabStyle = readMyPref("multizilla.session-manager.selectedtab-styling", "string", "bold");
    var mzRDF = "http://multizilla.mozdev.org/rdf#";
    var ordinal_property = RDF.GetResource(mzRDF + "ordinal");
    var selected_property = RDF.GetResource(mzRDF + "selected");
    var selectedTabStyle_property = RDF.GetResource(mzRDF + "selectedTabStyle");
    // Get the index of the selected tab
    for (i in resourceList) {
      resource = resourceList[i];
      // Is this the selected tab?
      if (mzRDFStore.GetTarget(RDF.GetResource(resource), selected_property, true) == "true") {
        selectedTab = i;
        break;
      }
    }
    // Check the 'ordinal', 'selected' and 'selectedTabStyle' properties
    for (i in resourceList) {
      dump("\nresourceList[i]: " + resourceList[i]);
      resource = RDF.GetResource(resourceList[i]);
      data = mzRDFStore.GetTarget(resource, ordinal_property, true);
      data = mzRDFStore.GetTarget(resource, ordinal_property, true);
      dump("\ndata: " + data);
      if (data) {
        currentValue = data.QueryInterface(Components.interfaces.nsIRDFLiteral);
        currentValue = currentValue.Value;
        dump("\nordinal: " + currentValue);
        // Is the current 'ordinal' property set right?
        if (currentValue != i)
          addOrChangeValue(mzRDFStore, resource, ordinal_property, RDF.GetLiteral(currentValue), RDF.GetLiteral(i));
      }
      else
        addOrChangeValue(mzRDFStore, resource, ordinal_property, null, RDF.GetLiteral(i));
      // Select the first tab when we didn't have one
      if (i == 0 && selectedTab == -1) {
        selectedTab = 0;
        currentValue = mzRDFStore.GetTarget(resource, selected_property, true);
        addOrChangeValue(mzRDFStore, resource, selected_property, currentValue, RDF.GetLiteral("true"));
        currentValue = mzRDFStore.GetTarget(resource, selectedTabStyle_property, true);
        addOrChangeValue(mzRDFStore, resource, selectedTabStyle_property, currentValue, RDF.GetLiteral(selectedTabStyle));
      }
      else { // We have a selected tab for this session
        data = mzRDFStore.GetTarget(resource, selected_property, true);
        // Is this the selected tab?
        dump("\ni == selectedTab:" + (i == selectedTab));
        if (i == selectedTab && data) {
          currentValue = data.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;
          dump("\nselected: " + currentValue);
          // Does the current value match the users preference?
          if (currentValue != selectedTabStyle)
            addOrChangeValue(mzRDFStore, resource, selectedTabStyle_property, data, RDF.GetLiteral(selectedTabStyle));
        }
        else { // No, not the selected tab
          var data = mzRDFStore.GetTarget(resource, selected_property, true);
          // Remove 'selected' property for all but the selected tab (when available)
          if (data)
            mzRDFStore.Unassert(resource, selected_property, data);

          data = mzRDFStore.GetTarget(resource, selectedTabStyle_property, true);
          // Remove 'selectedTabStyle' property for all but the selected tab (when available)
          if (data)
            mzRDFStore.Unassert(resource, selectedTabStyle_property, data);
        }
      }
    }
    mzFlushDataSource(mzRDFStore);
  },

  canDrop: function(aItemIndex, orientation)
  {
    // Prevent drops from other windows
    if (!sessionManager.dragResourceList)
      return false;
    // get target resource i.e. 'urn:session-N:window-N:tabs:root' or 'urn:session-N:window-N:tabs:tab-N'
    var resource = sessionManager.getResourceForRow(tree, aItemIndex); // urn:session-N:window-N:tabs:root

    if (this._isDropOnParentContainer(resource, orientation))
      return false;

    var isSessionContainer = (resource.match(/root/) == "root");

    switch(orientation)
    {
      case this.DROP_BEFORE:
           if (isSessionContainer) {
             if (aItemIndex == 0) // prevent drops before/above the first session container
               break;
           }
           else
             return this._prepareDragAction(resource, "insert", isSessionContainer);
           break;
      case this.DROP_ON:
           if (isSessionContainer)
             return this._prepareDragAction(resource, "append", isSessionContainer);
           break;
      case this.DROP_AFTER:
           var hasNextSibling = tree.treeBoxObject.view.hasNextSibling(aItemIndex, aItemIndex);
           var dragAction = (hasNextSibling) ? "insert" : "append";

           if (hasNextSibling)
             resource = sessionManager.getResourceForRow(tree, aItemIndex + 1);

           if (isSessionContainer)
             return this._prepareDragAction(resource, "insert", isSessionContainer);
           else
             return this._prepareDragAction(resource, dragAction, isSessionContainer);
    }
    this._prepareDragAction("", null, false);
    return false;
  },

  onDrop: function(aRow, orientation)
  {
    var resource = this.dragTarget;
    var targetSessionResource = (resource.indexOf('tab-') > 0) ? resource.replace(/tab-\w*/, 'root') : resource;
    this.dragTarget = (this.dragAction == "append") ? targetSessionResource : resource;
    var resourceList = mzGetResourcesFromContainer(mzRDFStore, targetSessionResource);
    var newTargetResource = targetSessionResource.replace(/tabs:root/, 'tabs:tab-');
    var i, tabDataObject, targetResource, propertyValues, hasSelectedTab, selectedTabStyle = "";
    var targetIndex = mzRDFCUtils.indexOf(mzRDFStore, RDF.GetResource(targetSessionResource), RDF.GetResource(this.dragTarget));
    var sessionsToBeValidated = new Array();
    var draggedItems = 0;

    if (this.dragAction == "append") {
      for (i in sessionManager.dragResourceList) {
        targetTabResource = newTargetResource + (resourceList.length + Number(i));
        resource = sessionManager.dragResourceList[i];
        propertyValues = getTabDataFromResource(mzRDFStore, RDF.GetResource(resource));
        var moveInParentContainer = (targetSessionResource == resource.replace(/tab-\w*/, 'root'));
        selectedTabStyle = (moveInParentContainer && propertyValues[SELECTED_PROPERTY] == "true") ? propertyValues[TABSTYLE_PROPERTY] : "";
        propertyValues[TABSTYLE_PROPERTY] = selectedTabStyle;
        mzAddOrInsertTabToSession(mzRDFStore, targetSessionResource, targetTabResource, 0, propertyValues, selectedTabStyle);

        if (moveInParentContainer)
          removeTabResourcesFor(mzRDFStore, resource);
        else if (this.dragSession.action & Components.interfaces.nsIDragService.DRAGDROP_ACTION_MOVE)
          removeTabResourcesFor(mzRDFStore, resource);

        mzFlushDataSource(mzRDFStore);
        draggedItems++;
      }
      // Do we have to remove items from RDF?
      if (draggedItems && (this.dragSession.action & Components.interfaces.nsIDragService.DRAGDROP_ACTION_MOVE)) {
        // Iterate over the dragged items
        for (i in sessionManager.dragResourceList) {
          // Is this item dragged to a different session container?
          if (sessionManager.dragResourceList[i].replace(/tab-\w*/, 'root') != targetSessionResource) {
            removeTabResourcesFor(mzRDFStore, sessionManager.dragResourceList[i]);
            mzFlushDataSource(mzRDFStore);
          }
        }
      }
      this._checkContainerIndex(targetSessionResource);
    }
    else if (this.dragAction == "insert") { // Also used for moving tabs in a session container!
      // Replace the resource strings with data objects from the corresponding items
      for (rIndex in resourceList) {
        tabDataObject = new Object();
        tabDataObject.resource = resourceList[rIndex];
        tabDataObject.propertyValues = getTabDataFromResource(mzRDFStore, RDF.GetResource(tabDataObject.resource));
        tabDataObject.isnew = false;
        resourceList[rIndex] = tabDataObject;
      }
      // Insert the dragged items into our data object array
      for (i = (sessionManager.dragResourceList.length - 1); i >= 0; i--) {
        tabDataObject = new Object();
        tabDataObject.resource = sessionManager.dragResourceList[i];
        tabDataObject.propertyValues = getTabDataFromResource(mzRDFStore, RDF.GetResource(tabDataObject.resource));
        tabDataObject.isnew = true;
        resourceList.splice((targetIndex-1), 0, tabDataObject); // Arrays start with 0 but index with 1
      }
      sessionManager.dragResourceList = null;
      // Remove all items from target session container
      var sessionsContainerItems = mzGetResourcesFromContainer(mzRDFStore, targetSessionResource);

      for (i in sessionsContainerItems) {
        removeTabResourcesFor(mzRDFStore, sessionsContainerItems[i], false);
      }
      // Remove dragged items from RDF and resourceList, when needed
      if (this.dragSession.action & Components.interfaces.nsIDragService.DRAGDROP_ACTION_MOVE) {
        for (rIndex in resourceList) {
          // Is this a newly inserted item?
          if (resourceList[rIndex].isnew) {
            resource = resourceList[rIndex].resource;
            // Lookup resource and remove it from RDF and data object array
            for (i in resourceList) {
              if (resourceList[i].isnew)
                continue;
              else if (resource == resourceList[i].resource) {
                resourceList.splice(i, 1);
                break;
              }
            }
            // Remove dragged items from the source session container
            if (resource.indexOf(newTargetResource) < 0) {
              removeTabResourcesFor(mzRDFStore, resource, false);
              // Did we remove the selected tab?
              if (resourceList[rIndex].propertyValues[SELECTED_PROPERTY] == "true")
                sessionsToBeValidated.push(resource.replace(/tab-\w*/, 'root'));
            }
          }
        }
      }
      // We might have dragged the selected tab out of a session 
      for (i in sessionsToBeValidated) {
        this._validateContainer(sessionsToBeValidated[i]);
      }
      /***
        * Do we (still) have a selected tab in the target session container? 
        * Note that there won't be one when your drag it to another session!
        */
      for (i in resourceList) {
        if (resourceList[i].propertyValues[SELECTED_PROPERTY] == "true") {
          hasSelectedTab = true;
          break;
        }
      }
      // Select the first tab when we didn't have one
      if (!hasSelectedTab) {
        selectedTabStyle = readMyPref("multizilla.session-manager.selectedtab-styling", "string", "bold");
        resourceList[0].propertyValues[TABSTYLE_PROPERTY] = selectedTabStyle;
        resourceList[0].propertyValues[SELECTED_PROPERTY] = "true";
        selectedTabStyle = "";
      }
      // Append all items from data object array
      for (i = 0; i < resourceList.length; i++) {
        targetResource = newTargetResource + i;
        propertyValues = resourceList[i].propertyValues;
        // Make sure we only have one selected tab, with optional tab styling
        if (selectedTabStyle == "" && propertyValues[SELECTED_PROPERTY] == "true")
          selectedTabStyle = propertyValues[TABSTYLE_PROPERTY];
        else {
          propertyValues[TABSTYLE_PROPERTY] = "";
          propertyValues[SELECTED_PROPERTY] = "false";
        }
        mzAddOrInsertTabToSession(mzRDFStore, targetSessionResource, targetResource, 0, propertyValues, selectedTabStyle);
        mzFlushDataSource(mzRDFStore);
      }
    }
    tree.builder.rebuild();
    sessionManager.dragResourceList = null;
    this.dragSession = null;
  },
  // Backward compatibility with older Mozilla versions/builds
  canDropOn: function(aRow)
  {
    var orientation = this.DROP_ON;
    return this.canDrop(aRow, orientation);
  },
  // Backward compatibility with older Mozilla versions/builds
  canDropBeforeAfter: function(aRow, before)
  {
    var orientation = (before) ? this.DROP_BEFORE : this.DROP_AFTER;
    return this.canDrop(aRow, orientation);
  },

  onToggleOpenState    : function (aRow)                           {},
  onCycleHeader        : function (aColumnID, aHeaderElement)      {},
  onSelectionChanged   : function ()                               {},
  onCycleCell          : function (aItemIndex, aColumnID)          {},
  isEditable           : function (aItemIndex, aColumnID)          {},
  onSetCellText        : function (aItemIndex, aColumnID, aValue)  {},
  onPerformAction      : function (aAction)                        {},
  onPerformActionOnRow : function (aAction, aItemIndex)            {},
  onPerformActionOnCell: function (aAction, aItemIndex, aColumnID) {}
}

addEventListener("load", sessionManager.initSessionManager, false);
