/* 
   - The contents of this file are subject to the Mozilla Public
   - License Version 1.1 (the "License"); you may not use this file
   - except in compliance with the License. You may obtain a copy of
   - the License at http://www.mozilla.org/MPL/
   -
   - Software distributed under the License is distributed on an "AS
   - IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
   - implied. See the License for the specific language governing
   - rights and limitations under the License.
   -
   - The Original Code is this file as it was released on
   - 26 January 2004.
   -
   - The Initial Developer of the Original Code is HJ van Rantwijk.
   - Portions created by HJ van Rantwijk are Copyright (C) 2004
   - HJ van Rantwijk.  All Rights Reserved.
   -
*/

const uaPrefIdentifier = "general.useragent.override";
const NO_ROWS_AVAILABLE      = -1;
const NO_ROWS_SELECTED       = 0;
const SINGLE_ROW_SELECTED    = 1;
const MULTIPLE_ROWS_SELECTED = 2;

var pref = null;
var prefWindow = null;
var prefPanelNames;
var managerTypes = [ "userAgentManager" ];
var dialogType = "userAgentManagerPrefWindow";
var panelPath = "chrome://multiviews/content/useragent-manager/preferences/";
var gUseStackedWindows = false;
var activeUserAgent = null;
// Global vars used by confirmation dialog
var gConfirmCheckbox = false;
var gEscapeKeyPressed = false;
var gConfirmForID = 0;

/***
  * Toolbar button ID arrays that will be replaced with elements in updateToolbarButtons()
  * note: keep the order in-line with the corresponding XUL code!
  */
var toolbarButtons = [ "originalButton", "customButton", "addButton", "replaceButton", "modifyButton", "deleteButton", "prefsButton" ];
/***
  * Context Menu ID array that will be replaced with elements in initContextMenu(aEvent)
  * note: keep the order in-line with the corresponding XUL code!
  */
var contextMenuItems = [ "selectAll", "undoSelection", "selectDefaultUA", "selectCustomUA", "addUA", "replaceUA", "editUA", "removeUA" ];
/***
  * Toolbar button arrays, to ease the pain of enabling/disabling the buttons
  */
var defaultEnabledButtons      = [ false, false, false, false, false, false, true ];
var selectionEnabledButtons    = [ false, false, false, false,  true,  true, true ];
var multipleRowsEnabledButtons = [ false, false, false, false, false,  true, true ];
var modifyEnabledButtons       = [ false, false,  true,  true, false, false, true ];

const DEFAULT_BUTTON = 0;
const CUSTOM_BUTTON  = 1;
const REPLACE_BUTTON = 3;

const CM_SELECT_ALL  = 0;
const CM_SELECT_NONE = 1;
const CM_DEFAULT     = 2;
const CM_CUSTOM      = 3;
const CM_ADD         = 4;
const CM_REPLACE     = 5;
const CM_EDIT        = 6;
const CM_REMOVE      = 7;

////////////////////////////////////////////////////////////////////////////////
function selectWindowStyling(aWindow)
{
  var documentElement = aWindow.document.documentElement;

  if (readMyPref("multizilla.windows.classic-style", "bool", false))
    documentElement.removeAttribute("extraStyling");
  else
    documentElement.setAttribute("extraStyling", "true");
}
////////////////////////////////////////////////////////////////////////////////
function getTabbrowser(aOpenFlag)
{
  var navigatorWindow = getNavigatorWindow(aOpenFlag);

  if (navigatorWindow)
    return navigatorWindow.getBrowser();
  return null;
}
////////////////////////////////////////////////////////////////////////////////
function readMyPref(aPrefString, aPrefType, aDefaultValue)
{
  var navigatorWindow = getNavigatorWindow(true);
  return navigatorWindow.mzPreferences.readMyPref(aPrefString, aPrefType, aDefaultValue);
}
////////////////////////////////////////////////////////////////////////////////
function updateToolbarButtons()
{
  // Is the toolbar button array initialized?
  if (typeof toolbarButtons[0] == 'string') {
    // Nope, not yet, initialize array now
    for (i in toolbarButtons) {
      toolbarButtons[i] = document.getElementById(toolbarButtons[i]);
    }
  }
  var skipCustomDefault = false;
  var buttons = defaultEnabledButtons;
  var treeState = userAgentManager.getCurrentTreeState();
  
  if (textField.value != "") {
    skipCustomDefault = true;
    buttons = modifyEnabledButtons;
    buttons[REPLACE_BUTTON] = textField.edit;
  }
  else if (treeState > NO_ROWS_SELECTED) {
    if (treeState == SINGLE_ROW_SELECTED)
      buttons = selectionEnabledButtons;
    else if (treeState == MULTIPLE_ROWS_SELECTED)
      buttons = multipleRowsEnabledButtons;
  }
  // Enable/disable the toolbar buttons based on the toolbar array
  for (i in toolbarButtons) {
    toolbarButtons[i].disabled = !buttons[i];
  }
  if (!skipCustomDefault) {
    // Enable/disable the 'Default' button on the toolbar
    toolbarButtons[DEFAULT_BUTTON].disabled = (activeUserAgent == null);
    // Enable/disable the 'Custom' button on the toolbar
    if (currentRow > -1) {
      var columnID = "userAgentColumn";
      columnID = ('columns' in tree) ? tree.columns[columnID] : columnID;
      var onActiveUA = (tree.builderView.getCellText(currentRow, columnID) == activeUserAgent);
      toolbarButtons[CUSTOM_BUTTON].disabled = onActiveUA || (treeState != SINGLE_ROW_SELECTED);
    }
  }
  var disabled = (currentUserAgent == activeUserAgent) || (treeState != SINGLE_ROW_SELECTED);
  document.documentElement.getButton("accept").disabled = disabled;
}
////////////////////////////////////////////////////////////////////////////////
function initContextMenu(aEvent)
{
  // Is the context menu item array initialized?
  if (typeof contextMenuItems[0] == 'string') {
    // Nope, not yet, initialize array now
    for (i in contextMenuItems) {
      contextMenuItems[i] = document.getElementById(contextMenuItems[i]);
    }
  }
  for (i in contextMenuItems) {
    contextMenuItems[i].setAttribute('disabled', 'true');
  }
  var treeState = userAgentManager.getCurrentTreeState();

  if (treeState == NO_ROWS_AVAILABLE)
    return;

  var treeRows = tree.treeBoxObject.view.rowCount;

  if (treeState == MULTIPLE_ROWS_SELECTED) {
    contextMenuItems[CM_SELECT_ALL].setAttribute('disabled', 'true');
    contextMenuItems[CM_SELECT_NONE].removeAttribute('disabled');
  }
  else { // NO_ROWS_SELECTED || SINGLE_ROW_SELECTED
    if (treeRows > 1 && textField.value == "") 
      contextMenuItems[CM_SELECT_ALL].removeAttribute('disabled');
    contextMenuItems[CM_SELECT_NONE].setAttribute('disabled', 'true');
  }
  var overrideUA;

  try {
    overrideUA = pref.getCharPref(uaPrefIdentifier);
  } catch(ex) { } // No User Agent override selected -> default UserAgent

  contextMenuItems[CM_DEFAULT].setAttribute('disabled', (activeUserAgent == null || textField.value != "") ? 'true' : 'false');

  if (treeState == NO_ROWS_SELECTED)
    return;

  contextMenuItems[CM_CUSTOM].setAttribute('disabled', (treeState == MULTIPLE_ROWS_SELECTED || activeUserAgent != null || textField.value != "") ? 'true' : 'false');
  contextMenuItems[CM_ADD].setAttribute('disabled', (textField.value == "") ? 'true' : 'false');
  contextMenuItems[CM_REPLACE].setAttribute('disabled', (textField.value == "" || textField.edit == false) ? 'true' : 'false');
  contextMenuItems[CM_EDIT].setAttribute('disabled', (treeState == MULTIPLE_ROWS_SELECTED || textField.value != "") ? 'true' : 'false');
  contextMenuItems[CM_REMOVE].setAttribute('disabled', (textField.value != "") ? 'true' : 'false');
}
////////////////////////////////////////////////////////////////////////////////
function openPrefOverlay()
{
  var winFeatures = "chrome, dependent, modal, dialog" + (gUseStackedWindows ? "" : ", resizable=yes, centerscreen");
  window.openDialog("chrome://multiviews/content/useragent-manager/preferences/userAgentManagerPrefWindow.xul", 
                    "", winFeatures, window, "");
}
////////////////////////////////////////////////////////////////////////////////
function switchPanel(aPanel)
{
  var categoryButton = null;
  var panelFile = null;

  prefWindow.gSecondLevelPrefs = false;

  if (aPanel == "userAgentManagerPrefWindow")
    panelFile = "userAgentManagerMainPanel.xul";
  prefWindow.refreshPanel(panelFile, categoryButton);
}
////////////////////////////////////////////////////////////////////////////////
function initWindowOpenType()
{
  try {
    gUseStackedWindows = pref.getBoolPref("multizilla.managers.use-stacked-windows");
  } catch(ex) {
    gUseStackedWindows = /Win/.test(navigator.platform);
  }
}
////////////////////////////////////////////////////////////////////////////////
function getConfirmation(aWindow, aTitle, aText, aCheckboxText)
{
  gConfirmCheckbox = false;
  var checkbox = {};
  var checkboxText = aCheckboxText ? stringBundle.getString(aCheckboxText) : null;
  var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
  var result = promptService.confirmEx(aWindow, stringBundle.getString(aTitle), stringBundle.getString(aText),
                                       (promptService.BUTTON_POS_0 * promptService.BUTTON_TITLE_YES) + 
                                       (promptService.BUTTON_POS_1 * promptService.BUTTON_TITLE_NO),
                                       null, null, null, checkboxText, checkbox);

  gConfirmCheckbox = checkbox.value;
  return (result == 0); // Yes returns 0 / No, Esc and X returns 1
}
////////////////////////////////////////////////////////////////////////////////
var userAgentManager = 
{
  tree : null,
  uaEntries : null,
  currentRow : null,
  tabbrowser : null,
  textField : null,
  stringBundle : null,
  currentUserAgent : null,
  dragStartRow : null,

  initUserAgentManager: function()
  {
    tree = document.getElementById('userAgentManagerTree');
    mzRDFStore = mzGetNewOrCurrentDSByFilename("useragents.rdf", true, null, false);
    tree.database.AddDataSource(mzRDFStore);
    userAgentManager.doInit();
    stringBundle = document.getElementById('properties');
    textField = document.getElementById('newUserAgentField');
    textField.edit = false;
    currentUserAgent = null;
    currentRow = -1;
    dragStartRow = null;
    tabbrowser = getTabbrowser(true);

    var prefService = Components.classes["@mozilla.org/preferences-service;1"]
                                .getService(Components.interfaces.nsIPrefService);
    pref = prefService.getBranch(null);
    prefPanelNames = stringBundle.getString("prefPanelNames").split(',');

    if (pref.prefHasUserValue(uaPrefIdentifier))
      activeUserAgent = pref.getCharPref(uaPrefIdentifier);

    setTimeout(userAgentManager.selectRow, 0);
    sizeToContent();
    initWindowOpenType();
    updateToolbarButtons();

    if (tabbrowser.getPrefInitOnFailure("multizilla.uam.show-help-onstartup", "bool", true))
      userAgentManager.visitHelpPage();
  },

  doInit: function()
  {
    tree.builder.rebuild();
    uaEntries = this.initEntries();
  },

  getRowNumber: function(aEvent)
  {
    var X = aEvent.pageX;
    var Y = aEvent.pageY;
    return tree.treeBoxObject.getRowAt(X, Y);
  },

  getSelectedRows: function()
  {
    var rows = -1;
    var selectedRows = new Array();
    // Get and return all selection ranges
    var rangeCount = tree.builderView.selection.getRangeCount();
    // Iterate over all ranges to collect the selected rows
    for (var range = 0; range < rangeCount; range++) {
      var startSelection = {};
      var endSelection = {};
      tree.builderView.selection.getRangeAt(range, startSelection, endSelection);
      // Collect all rows and store them in our array
      for (var row = startSelection.value; row <= endSelection.value; row++) {
        // Check for invalid row numbers
        if (startSelection.value != -1 || endSelection.value != -1)
          selectedRows[++rows] = row;
      }
    }
    if (selectedRows)
      return selectedRows;
    return null;
  },

  getCurrentTreeState: function()
  {
    if (tree.treeBoxObject.view.rowCount < 1)
      return NO_ROWS_AVAILABLE;

    var selectedRows = this.getSelectedRows().length;

    if (selectedRows > 1)
      return MULTIPLE_ROWS_SELECTED;
    else if (selectedRows == 1)
      return SINGLE_ROW_SELECTED;
    return NO_ROWS_SELECTED;
  },

  selectRow: function()
  {
    var rows = tree.treeBoxObject.view.rowCount;

    for (var i = 0; i < rows; i++) {
      var columnID = "userAgentColumn";
      columnID = ('columns' in tree) ? tree.columns[columnID] : columnID;
      var userAgent = tree.builderView.getCellText(i, columnID);

      if (userAgent == navigator.userAgent) {
        currentRow = i;
        tree.builderView.selection.toggleSelect(i);
        tree.treeBoxObject.ensureRowIsVisible(i);
        tree.focus();
        break;
      }
    }
  },
  
  initEntries: function()
  {
    var entries = null;
    var RDFC = Components.classes["@mozilla.org/rdf/container-utils;1"].getService(Components.interfaces.nsIRDFContainerUtils);
    entries = RDFC.MakeSeq(mzRDFStore, RDF.GetResource("urn:useragent-data"));

    if (!entries)
      return null;
    if (!entries.GetElements())
      return null;
    return entries;
  },

  moveEntries: function(sourceRow, targetRow)
  {
    if (sourceRow > 0 && targetRow > 0 && (sourceRow != targetRow)) {
      var columnID = "userAgentColumn";
      columnID = ('columns' in tree) ? tree.columns[columnID] : columnID;
      var userAgent = tree.builderView.getCellText(sourceRow, columnID);
      uaEntries.RemoveElementAt(sourceRow, true);
      uaEntries.InsertElementAt(RDF.GetLiteral(userAgent), targetRow, true);
      this.doFlush();
      this.doInit();
    }
  },

  doAdd: function(addFlag)
  {
    var targetRow = (currentRow + 1);
    currentUserAgent = textField.value;

    if (!uaEntries || currentUserAgent == "")
      return;
    // Is this a duplicated entry?
    if (uaEntries.IndexOf(RDF.GetLiteral(currentUserAgent)) > 0) {
      var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
      promptService.alert(window, stringBundle.getString('uamDuplicatedItemTitle'), stringBundle.getString('uamDuplicatedItemText'));
      return;
    }
    if (addFlag) {
      var targetPosition = tabbrowser.getPrefInitOnFailure("multizilla.uam.insert-new-items-at", "int", 3);

      switch(targetPosition) {
        case 0: targetRow = 1; // insert new items at start of tree
                break;
        case 1: targetRow = (currentRow > 0) ? (currentRow + 1) : 1; // insert new items before selected item in tree
                break;
        case 2: targetRow++; // insert new items after selected item in tree
                break;
        default: targetRow = (tree.treeBoxObject.view.rowCount + 1); // add new items at end of tree
      }
      uaEntries.InsertElementAt(RDF.GetLiteral(currentUserAgent), targetRow, true);

      if (targetRow > 1)
        targetRow--;
    }
    else if ('edit' in textField && textField.edit) {
      uaEntries.RemoveElementAt(targetRow, true);
      uaEntries.InsertElementAt(RDF.GetLiteral(currentUserAgent), targetRow, true);
      targetRow--;
    }
    this.doFlush();
    this.doInit();
    textField.value = "";
    textField.edit = false;
    updateToolbarButtons();
    tree.builderView.selection.select(targetRow);
    tree.treeBoxObject.ensureRowIsVisible(targetRow);
    tree.focus();
  },

  doEdit: function()
  {
    if (currentRow > -1) {
      textField.value = currentUserAgent;
      textField.edit = true;
      updateToolbarButtons();
      textField.focus();
    }
  },

  doRemove: function(skipLastResourceCheck)
  {
    var rows = this.getSelectedRows();

    if (rows.length) {
      var doRemove = -1;
      var confirmActionText = (rows.length > 1) ? 'uamRemoveMultipleItems' : 'uamRemoveSingleItem';
      var removeCurrentCheck = tabbrowser.getPrefInitOnFailure("multizilla.uam.prevent-delete-current", "bool", true);

      if (rows.length == 1 && removeCurrentCheck) {
        doRemove = this.checkCurrentRemoval(rows[0]);

        if (doRemove == false)
          return;
      }
      if (doRemove == -1 && !this.doConfirm(confirmActionText))
        return;

      for (var i = rows.length-1; i > -1; i--) {
        var row = (rows.length > 1) ? rows[i] : currentRow;

        if (removeCurrentCheck && doRemove == -1 && this.checkCurrentRemoval(rows[i]) == false)
          continue;
        try {
          var uaItems = uaEntries.GetElements();
          uaEntries.RemoveElementAt((row + 1), true);
        } catch(ex) {
          var removeItemFailed = stringBundle.getString('uamRemoveItemFailed');
          alert(removeItemFailed);
        }
      }
      this.doFlush();
      this.doInit();
    }
    updateToolbarButtons();
  },

  checkCurrentRemoval: function(targetRow)
  {
    var columnID = "userAgentColumn";
    columnID = ('columns' in tree) ? tree.columns[columnID] : columnID;

    if (tree.builderView.getCellText(targetRow, columnID) == activeUserAgent)
      return this.doConfirm('uamRemoveCurrent');
    return -1;
  },

  doFlush: function()
  {
     mzRDFStore.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource).Flush();
  },

  doSelect: function(aCommand)
  {
    switch (aCommand) {
      case "all": // Called from the context menu, or Ctrl+A ?
           tree.builderView.selection.selectAll();
           break;
      case "none": // Called from the context menu, or Ctrl+Z ?
           tree.builderView.selection.clearSelection();
           tree.builderView.selection.select(tree.currentIndex);
           break;
    }
    tree.focus();
  },

  doConfirm: function(additionalText)
  {
    var doAction = false;
    var dialogTitle = stringBundle.getString('uamConfirmActionTitle');
    var dialogText = stringBundle.getString('uamConfirmActionText');
    additionalText = stringBundle.getString(additionalText);
    dialogText = dialogText.replace(/%S/, additionalText);
    var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService();

    promptService = promptService.QueryInterface(Components.interfaces.nsIPromptService);
    doAction = promptService.confirm(window, dialogTitle, dialogText);
    tree.focus();
    return doAction;
  },

  doActivate: function(selectDefault, closeManager)
  {
    if (selectDefault) {
      pref.setCharPref(uaPrefIdentifier, "");
      pref.clearUserPref(uaPrefIdentifier);
      activeUserAgent = null;
    }
    else if (currentUserAgent) {
      if (userAgentManager.getSelectedRows().length > 1)
        return;
      pref.setCharPref(uaPrefIdentifier, currentUserAgent);
      activeUserAgent = currentUserAgent;
      // tree.builderView.getResourceAtIndex(tree.currentIndex).Value); //.setAttribute("properties", "activated");
      // tree.contentView.getElementAtIndex(tree.currentIndex);
    }
    textField.value = "";
    updateToolbarButtons();

    var aboutShow = tabbrowser.getPrefInitOnFailure("multizilla.uam.show-about-page", "bool", true);
    var aboutClose = tabbrowser.getPrefInitOnFailure("multizilla.uam.close-about-page", "bool", true);

    if (aboutShow && !(closeManager && aboutClose))
      tabbrowser.selectedTab = tabbrowser.addOrGetTab("about:");
    if (closeManager)
      this.removeAboutTab();
    tree.focus();
  },

  onTextFieldKeypress: function(aEvent)
  {
    aEvent.preventBubble();

    if (aEvent.keyCode == 13)
      userAgentManager.doAdd(!textField.edit);
    else if (aEvent.keyCode == 27) {
      textField.value = "";
      textField.edit = false;
      updateToolbarButtons();
      tree.focus();
    }
  },

  onClick: function(aEvent)
  {
    // if (aEvent.button != 2 && !aEvent.ctrlKey)
      // this.onSelect();
  },

  onSelect: function()
  {
    dump("\nonSelect()");
    currentRow = tree.currentIndex;

    if (currentRow > -1) {
      var columnID = "userAgentColumn";
      columnID = ('columns' in tree) ? tree.columns[columnID] : columnID; 
      currentUserAgent = tree.builderView.getCellText(currentRow, columnID);
    }
    textField.value = "";
    textField.edit = false;
    updateToolbarButtons();
  },

  visitHelpPage: function()
  {
    var helpPages = stringBundle.getString('helpPages-UserAgentManager');
    tabbrowser.selectedTab = tabbrowser.addOrGetTab(helpPages);
  },

  onHelp: function(aPageAnchor)
  {
    var pageAnchor = "";
    var helpPages = stringBundle.getString("helpPages-UserAgentManager");

    if (aPageAnchor) {
      if (typeof aPageAnchor == 'string')
        pageAnchor = "#" + aPageAnchor;
      else // event
        pageAnchor = "#" + aPageAnchor.target.id.replace(/^userAgentManager-/, '');
    }
    // We could eventually check the on-line state!
    tabbrowser = getTabbrowser(true);
    tabbrowser.selectedTab = tabbrowser.addOrGetTab(helpPages + pageAnchor);
    window.close();
  },

  onCancel: function()
  {
    var shouldClose = this.confirmClose(gConfirmForID);

    if (shouldClose && gConfirmForID == 1) { // Called from Close Button
      this.removeAboutTab();
    }
    return shouldClose;
  },

  onAccept: function()
  {
    this.doActivate(false, false);
    return false; // don't close this dialog window!
  },

  confirmClose: function(aConfirm)
  {
    var shouldClose = true;
    var confirmPrefID = "multizilla.ua-manager.confirm.quit";
    var dialogCheckboxText, confirmPrefValue = 31;
    /* ::::
       1  = Close Button
       2  = ESC key pressed
       4  = Ctrl-W
       8  = Click on [X] on Window
    :::: */
    try {
      // try to read the pref
        confirmPrefValue = pref.getIntPref(confirmPrefID);
    } catch(ex) {
      // Yikes, no pref found, add it with the default value
      pref.setIntPref(confirmPrefID, confirmPrefValue);
    }
    // Is the pref set for confirmation?
    if (confirmPrefValue && (confirmPrefValue & aConfirm) == aConfirm) {
      // Yes, ask the users confirmation
      switch(aConfirm) {
        case (aConfirm & 1): dialogCheckboxText = 'managerSkipConfirmForCloseButton';
          break;
        case (aConfirm & 2): dialogCheckboxText = 'managerSkipConfirmForEscapeKey';
          break;
        case (aConfirm & 4): dialogCheckboxText = 'managerSkipConfirmForCtrlW';
          break;
        case (aConfirm & 8): dialogCheckboxText = 'managerSkipConfirmForWindowX';
          break;
      }
      shouldClose = getConfirmation(window, 'tmConfirmCloseTitle', 'uaConfirmCloseText', dialogCheckboxText);
      // Is the checkbox checked and do we have a pref value to work with?
      if (gConfirmCheckbox && (confirmPrefValue >= aConfirm)) {
        // Yes, update pref
        pref.setIntPref(confirmPrefID, (confirmPrefValue -= aConfirm));
      }
    }
    return shouldClose;
  },

  removeAboutTab: function()
  {
     if (tabbrowser.getPrefInitOnFailure("multizilla.uam.close-about-page", "bool", true)) {
       var browsers = tabbrowser.mPanelContainer.childNodes.length;

       for (var i = 0; i < browsers; i++) {
         if (tabbrowser.mTabContainer.childNodes[i].hasAttribute("busy"))
           continue;
         if (tabbrowser.getBrowserAtIndex(i).webNavigation.currentURI.spec == "about:")
           tabbrowser.removeTab(tabbrowser.mTabContainer.childNodes[i]);
       }
     }
  },

  updateTextField: function updateTextField(aFlag)
  {
    // var addButton = document.getElementById('addButton');

    if (aFlag)
      textField.value = "";
    /* if (textField.value == "")
      addButton.setAttribute("disabled", "true");
    else
      addButton.removeAttribute("disabled"); */
    updateToolbarButtons();
  },

  onDragStart: function(aEvent, aXferData, aDragSession)
  {
    if (tree.treeBoxObject.view.rowCount > 0) {
      // aEvent.target.setAttribute("dragover", "true");
      var X = aEvent.pageX;
      var Y = aEvent.pageY;
      dragStartRow = tree.treeBoxObject.getRowAt(X, Y);
      var columnID = "userAgentColumn";
      columnID = ('columns' in tree) ? tree.columns[columnID] : columnID;
      currentUserAgent = tree.builderView.getCellText(dragStartRow, columnID);

      if (dragStartRow > -1) {
        aXferData.data = new TransferData();
        aXferData.data.addDataForFlavour("text/unicode", currentUserAgent);
      }
    }
  },

  onDragOver: function(aEvent, aXferData, aDragSession)
  {
    var X = aEvent.pageX;
    var Y = aEvent.pageY;
    var row = tree.treeBoxObject.getRowAt(X, Y);

    if (row > -1)
      aDragSession.canDrop = true;
  },

  onDrop: function(aEvent, aXferData, aDragSession)
  {
  },

  onDragExit: function(aEvent, aXferData, aDragSession)
  {
    var X = aEvent.pageX;
    var Y = aEvent.pageY;
    var row = tree.treeBoxObject.getRowAt(X, Y);

    if (row > -1) {
      this.moveEntries((dragStartRow + 1), (row + 1));
    }
    tree.childNodes[1].removeAttribute("dragover");
  },

  getSupportedFlavours: function ()
  {
    var flavourSet = new FlavourSet();
    flavourSet.appendFlavour("text/unicode");

    return flavourSet;
  },

  canDrop: function (aEvent, aDragSession)
  {
    aDragSession.canDrop = true;
    return true;
    // Allow drops from within User Agent Manager only!
    var doc = aDragSession.sourceNode.ownerDocument;
    var elem = doc.getElementById("userAgentManagerWindow");
    return true; // (elem && (elem.getAttribute("windowtype") == "multizilla:useragent-manager"));
  }
};

addEventListener("load", userAgentManager.initUserAgentManager, false);